/*****************************************************************************\
*              efs - General purpose Embedded Filesystem library              *
*          --------------------- -----------------------------------          *
*                                                                             *
* Filename :  efs_ls.c                                                        *
* Description : This file contains functions to list the files in a directory *
*                                                                             *
* This library is free software; you can redistribute it and/or               *
* modify it under the terms of the GNU Lesser General Public                  *
* License as published by the Free Software Foundation; either                *
* version 2.1 of the License, or (at your option) any later version.          *
*                                                                             *
* This library is distributed in the hope that it will be useful,             *
* but WITHOUT ANY WARRANTY; without even the implied warranty of              *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           *
* Lesser General Public License for more details.                             *
*                                                                             *
*                                                    (c)2005 Michael De Nil   *
*                                                    (c)2005 Lennart Yseboodt *
\*****************************************************************************/

/*****************************************************************************/
#include "efs_dir.h"
#include "efs_fat.h"
#include "efs_plibc.h"
#include "efs_ls.h"
/*****************************************************************************/

esint8 ls_openDir(DirList *dlist, FileSystem *fs, eint8 *dirname)
{
	FileLocation loc;
	euint32 fc;

	dlist->fs = fs;

	if (fs_findFile(dlist->fs, dirname, &loc, &fc) != 2)
	{
		return -1;
	}

	fs_initClusterChain(dlist->fs, &(dlist->Cache), fc);
	memClr(&(dlist->currentEntry), sizeof(dlist->currentEntry));
	dlist->rEntry = 0;
	dlist->cEntry = ~0;

	return 0;
}
/*****************************************************************************/

static esint8 ls_isValidFileEntry(ListDirEntry *entry)
{
	euint8 first_char = entry->FileName[0];

	if (	first_char < 0x20	/* invalid first character */
		 || first_char == 0xE5	/* deleted file */
		 || first_char == '.')	/* . or .. entry */
		return 0;
	if ((entry->Attribute & 0x0F) == 0x0F)	/* part of long filename */
		return 0;
	/* Only the Volume label can contain spaces */
	if (((entry->Attribute & ATTR_VOLUME_ID) == 0) && (first_char == ' '))
		return 0;

	return 1;
}
/*****************************************************************************/

void ls_fileEntryToDirListEntry(DirList *dlist, euint8 *buf, euint16 offset)
{
	if (offset > 480 || offset % 32)
		return;

	buf += offset;
	memCpy(buf + OFFSET_DE_FILENAME, dlist->currentEntry.FileName, LIST_MAXLENFILENAME);
	dlist->currentEntry.Attribute = *(buf + OFFSET_DE_ATTRIBUTE);
	dlist->currentEntry.FileSize = ex_getb32(buf, OFFSET_DE_FILESIZE);
	dlist->currentEntry.FileTime = ex_getb16(buf, OFFSET_DE_WRITETIME);
	dlist->currentEntry.FileDate = ex_getb16(buf, OFFSET_DE_WRITEDATE);
}
/*****************************************************************************/

static esint8 ls_getRealDirEntry(DirList *dlist)
{
	euint8 *buf;

	if (dlist->Cache.FirstCluster <= 1)
		return -1;

	if (fat_LogicToDiscCluster(dlist->fs,
							   &(dlist->Cache),
							   ((euint32)dlist->rEntry)
							   / (16 * (euint32)dlist->fs->volumeId.SectorsPerCluster)))
	{
		return -1;
	}

	buf = part_getSect(dlist->fs->part,
					   fs_clusterToSector(dlist->fs, dlist->Cache.DiscCluster)
					   + ((euint32)dlist->rEntry / 16) % dlist->fs->volumeId.SectorsPerCluster,
					   IOM_MODE_READONLY);

	if (buf == (euint8 *)0)
		return -1;

	ls_fileEntryToDirListEntry(dlist, buf, 32 * (dlist->rEntry % 16));
	part_relSect(dlist->fs->part, buf);

	return 0;
}
/*****************************************************************************/

static esint8 ls_getRootAreaEntry(DirList *dlist)
{
	euint8 *buf = 0;

	if ((dlist->fs->type != FAT12) && (dlist->fs->type != FAT16))
		return -1;
	if (dlist->rEntry >= dlist->fs->volumeId.RootEntryCount)
		return -1;

	buf = part_getSect(dlist->fs->part,
					   dlist->fs->FirstSectorRootDir + (euint32)dlist->rEntry / 16,
					   IOM_MODE_READONLY);
	if (buf == (euint8 *)0)
		return -1;

	ls_fileEntryToDirListEntry(dlist, buf, 32 * (dlist->rEntry % 16));
	part_relSect(dlist->fs->part, buf);
	return 0;
}
/*****************************************************************************/

static esint8 ls_getDirEntry(DirList *dlist)
{
	if (dlist->Cache.FirstCluster == 1)
	{
		return ls_getRootAreaEntry(dlist);
	}
	else if (dlist->Cache.FirstCluster)
	{
		return ls_getRealDirEntry(dlist);
	}
	return -1;
}
/*****************************************************************************/

esint8 ls_getNext(DirList *dlist)
{
	do
	{
		if (ls_getDirEntry(dlist))
			return -1;
		dlist->rEntry++;
	} while (!ls_isValidFileEntry(&(dlist->currentEntry)));
	dlist->cEntry++;
	return 0;
}
/*****************************************************************************/
