/*****************************************************************************\
*              efs - General purpose Embedded Filesystem library              *
*          --------------------- -----------------------------------          *
*                                                                             *
* Filename : sd.c                                                             *
* Revision : Initial developement                                             *
* Description : This file contains the functions needed to use efs for        *
*               accessing files on an SD-card.                                *
*                                                                             *
* This library is free software; you can redistribute it and/or               *
* modify it under the terms of the GNU Lesser General Public                  *
* License as published by the Free Software Foundation; either                *
* version 2.1 of the License, or (at your option) any later version.          *
*                                                                             *
* This library is distributed in the hope that it will be useful,             *
* but WITHOUT ANY WARRANTY; without even the implied warranty of              *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           *
* Lesser General Public License for more details.                             *
*                                                                             *
*                                                    (c)2005 Michael De Nil   *
*                                                    (c)2005 Lennart Yseboodt *
*                                     sd_getDriveSize added by Martin Thomas  *
*                    support for High Capacity SD cards added by Willi Thei  *
\*****************************************************************************/

/*****************************************************************************/
#include "interfaces/sd.h"
#include "efs_debug.h"
/*****************************************************************************/

/* macro for avoiding warnings about unused function arguments */
#define NOT_USED(x)		(void)(x)

/* MMC byte codes */
#define MMC_BC_CMD_ACK	0x00
#define MMC_BC_RST_ACK	0x01
#define MMC_BC_START	0xFE
#define MMC_BC_READY	0xFF

/* bits in R1 response */
#define MMC_R1_IN_IDLE		0x01
#define MMC_R1_ERASE_RESET	0x02
#define MMC_R1_ILL_CMD		0x04
#define MMC_R1_COM_CRC		0x08
#define MMC_R1_ERASE_SEQ	0x10
#define MMC_R1_ADDR_ERR		0x20
#define MMC_R1_PARAM_ERR	0x40
#define MMC_R1_VALID_BIT	0x80

/* check pattern for High Capacity card test */
#define SD_CHECK_PATTERN	0x000001AA

#define SD_HIGH_CAPACITY	(1u << 30)	/* CCS == Card Capacity Status */
#define SD_STD_CAPACITY		0

/*
	MMC commands
	[MSB=start bit (0), MSB-1=transmission bit (1), MSB-2 .. LSB = cmd
 */
#define MMC_CMD_TRANS_BIT				(1 << 6)

/* Class 0 & 1 commands: Basic commands and Read Stream commands */
#define MMC_CMD_GO_IDLE_STATE			( 0 | MMC_CMD_TRANS_BIT)
#define MMC_CMD_SEND_OP_COND			( 1 | MMC_CMD_TRANS_BIT)
#define MMC_CMD_ALL_SEND_CID			( 2 | MMC_CMD_TRANS_BIT)
#define MMC_CMD_SET_RELATIVE_ADDR		( 3 | MMC_CMD_TRANS_BIT)
#define MMC_CMD_SET_DSR					( 4 | MMC_CMD_TRANS_BIT)
#define MMC_CMD_SEL_DESEL_CARD			( 7 | MMC_CMD_TRANS_BIT)
#define MMC_CMD_SEND_CSD				( 9 | MMC_CMD_TRANS_BIT)
#define MMC_CMD_SEND_CID				(10 | MMC_CMD_TRANS_BIT)
#define MMC_CMD_STOP_TRANSMISSION		(12 | MMC_CMD_TRANS_BIT)
#define MMC_CMD_SEND_STATUS				(13 | MMC_CMD_TRANS_BIT)
#define MMC_CMD_GO_INACTIVE_STATE		(15 | MMC_CMD_TRANS_BIT)

/* Class 2 commands: Block oriented Read commands */
#define MMC_CMD_SET_BLOCKLEN			(16 | MMC_CMD_TRANS_BIT)
#define MMC_CMD_READ_SINGLE_BLOCK		(17 | MMC_CMD_TRANS_BIT)
#define MMC_CMD_READ_MULTIPLE_BLOCK		(18 | MMC_CMD_TRANS_BIT)

/* Class 3 commands: Sequential Write commands */
#define MMC_CMD_WRITE_DAT_UNTIL_STOP	(20 | MMC_CMD_TRANS_BIT)

/* Class 4 commands: Block oriented Write commands */
#define MMC_CMD_WRITE_BLOCK				(24 | MMC_CMD_TRANS_BIT)
#define MMC_CMD_WRITE_MULTIPLE_BLOCK	(25 | MMC_CMD_TRANS_BIT)
#define MMC_CMD_PROGRAM_CSD				(27 | MMC_CMD_TRANS_BIT)

/* Class 6 commands: Group Write protect commands */
#define MMC_CMD_SET_WRITE_PROT			(28 | MMC_CMD_TRANS_BIT)
#define MMC_CMD_CLR_WRITE_PROT			(29 | MMC_CMD_TRANS_BIT)
#define MMC_CMD_SEND_WRITE_PROT			(30 | MMC_CMD_TRANS_BIT)

/* Class 5 commands: Erase commands */
#define MMC_CMD_ERASE_WR_BLK_START		(32 | MMC_CMD_TRANS_BIT)
#define MMC_CMD_ERASE_WR_BLK_END		(33 | MMC_CMD_TRANS_BIT)
#define MMC_CMD_UNTAG_SECTOR			(34 | MMC_CMD_TRANS_BIT)
#define MMC_CMD_TAG_ERASE_GROUP_START	(35 | MMC_CMD_TRANS_BIT)
#define MMC_CMD_TAG_ERASE_GROUP_END		(36 | MMC_CMD_TRANS_BIT)
#define MMC_CMD_UNTAG_ERASE_GROUP		(37 | MMC_CMD_TRANS_BIT)
#define MMC_CMD_ERASE					(38 | MMC_CMD_TRANS_BIT)

/* Class 7 commands: Lock commands */
#define MMC_CMD_LOCK_UNLOCK				(42 | MMC_CMD_TRANS_BIT)

/* Class 8 commands: Application specific commands */
#define MMC_CMD_APP_CMD					(55 | MMC_CMD_TRANS_BIT)
#define MMC_CMD_GEN_CMD					(56 | MMC_CMD_TRANS_BIT)

#define SDCARD_CMD_STATUS				(13 | MMC_CMD_TRANS_BIT)
#define SDCARD_CMD_APP_OP_COND			(41 | MMC_CMD_TRANS_BIT)
#define SDCARD_CMD_SEND_SCR				(51 | MMC_CMD_TRANS_BIT)

/* Class ? commands: xxx commands */
#define MMC_CMD_READ_OCR				(58 | MMC_CMD_TRANS_BIT)
#define MMC_CMD_CRC_ON_OFF				(59 | MMC_CMD_TRANS_BIT)

/* High Capacity SD card commands */
#define SDCARD_CMD_SEND_IF_COND			( 8 | MMC_CMD_TRANS_BIT)


/* remark: data got from SPI interfaces must first be byte reversed */
/* CSD (Card Specific Data) structure */
typedef union
{
	/* CSD layout for MultiMedia cards */
	struct csd_mmc_s
	{
		/* bits 96 - 127 */
		unsigned tran_speed        : 8;	/* max. data transfer rate */
		unsigned nsac              : 8;	/* data read access-time-2 in CLK cycles (NSAC*100) */
		unsigned taac              : 8;	/* data read access-time-1 */
		unsigned                   : 2;
		unsigned spec_vers         : 4;	/* Spec version */
		unsigned csd_structure     : 2;	/* CSD structure */

		/* bits 64 - 95 */
		unsigned c_size            : 10;/* Device size (bit 2-11) */
		unsigned                   : 2;
		unsigned dsr_imp           : 1;	/* DSR implemented */
		unsigned read_blk_misalign : 1;	/* Read block misalignment */
		unsigned write_blk_misalign: 1;	/* Write block misalignment */
		unsigned read_blk_partial  : 1;	/* Partial blocks for read allowed */
		unsigned read_blk_len      : 4;	/* Max. read data block length */
		unsigned ccc               : 12;/* Card command classes */

		/* bits 32 - 63 */
		unsigned wp_grp_size       : 5;	/* Write protect group size */
		unsigned erase_grp_mult    : 5;	/* Erase group size multiplier */
		unsigned erase_grp_size    : 5;	/* Erase group size */
		unsigned c_size_mult       : 3;	/* Device size multiplier */
		unsigned vdd_w_curr_max    : 3;	/* Max. write current at VDD max */
		unsigned vdd_w_curr_min    : 3;	/* Max. write current at VDD min */
		unsigned vdd_r_curr_max    : 3;	/* Max. read current at VDD max */
		unsigned vdd_r_curr_min    : 3;	/* Max. read current at VDD min */
		unsigned c_size01          : 2;	/* Device size (bit 0-1) */

		/* bits 0 - 31 */
		unsigned                   : 1;	/* not used, always 1 */
		unsigned crc               : 7;	/* 7-bit CRC */
		unsigned ecc               : 2;	/* ECC code */
		unsigned file_format       : 2;	/* File format */
		unsigned tmp_wp            : 1;	/* Temporary write protection */
		unsigned perm_wp           : 1;	/* Permanent write protection */
		unsigned copy              : 1;	/* Copy flag (OTP) */
		unsigned file_format_grp   : 1;	/* File format group */
		unsigned                   : 5;
		unsigned write_blk_partial : 1;	/* Partial blocks for write allowed */
		unsigned write_blk_len     : 4;	/* Max. write data length */
		unsigned r2w_factor        : 3;	/* Write speed factor */
		unsigned default_ecc       : 2;	/* Manufacturer default ECC */
		unsigned wp_grp_enable     : 1;	/* Write protect group enable */
	} mmc;

	/* CSD layout for SD cards (standard capacity) */
	struct csd_sd_s
	{
		/* bits 96 - 127 */
		unsigned tran_speed        : 8;	/* max. data transfer rate */
		unsigned nsac              : 8;	/* data read access-time-2 in CLK cycles (NSAC*100) */
		unsigned taac              : 8;	/* data read access-time-1 */
		unsigned                   : 6;
		unsigned csd_structure     : 2;	/* CSD structure */

		/* bits 64 - 95 */
		unsigned c_size            : 10;/* Device size (bit 2-11) */
		unsigned                   : 2;
		unsigned dsr_imp           : 1;	/* DSR implemented */
		unsigned read_blk_misalign : 1;	/* Read block misalignment */
		unsigned write_blk_misalign: 1;	/* Write block misalignment */
		unsigned read_blk_partial  : 1;	/* Partial blocks for read allowed */
		unsigned read_blk_len      : 4;	/* Max. read data block length */
		unsigned ccc               : 12;/* Card command classes */

		/* bits 32 - 63 */
		unsigned wp_grp_size       : 7;	/* Write protect group size */
		unsigned sector_size       : 7;	/* Erase sector size */
		unsigned erase_blk_en      : 1;	/* Erase single block enable */
		unsigned c_size_mult       : 3;	/* Device size multiplier */
		unsigned vdd_w_curr_max    : 3;	/* Max. write current at VDD max */
		unsigned vdd_w_curr_min    : 3;	/* Max. write current at VDD min */
		unsigned vdd_r_curr_max    : 3;	/* Max. read current at VDD max */
		unsigned vdd_r_curr_min    : 3;	/* Max. read current at VDD min */
		unsigned c_size01          : 2;	/* Device size (bit 0-1) */

		/* bits 0 - 31 */
		unsigned                   : 1;	/* not used, always 1 */
		unsigned crc               : 7;	/* 7-bit CRC */
		unsigned                   : 2;
		unsigned file_format       : 2;	/* File format */
		unsigned tmp_wp            : 1;	/* Temporary write protection */
		unsigned perm_wp           : 1;	/* Permanent write protection */
		unsigned copy              : 1;	/* Copy flag (OTP) */
		unsigned file_format_grp   : 1;	/* File format group */
		unsigned                   : 5;
		unsigned write_blk_partial : 1;	/* Partial blocks for write allowed */
		unsigned write_blk_len     : 4;	/* Max. write data block length */
		unsigned r2w_factor        : 3;	/* Write speed factor */
		unsigned                   : 2;
		unsigned wp_grp_enable     : 1;	/* Write protect group enable */
	} sd;

	/* CSD layout for High Capacity SD cards (version 2.0) */
	struct csd_sd_v2_s
	{
		/* bits 96 - 127 */
		unsigned tran_speed        : 8;	/* max. data transfer rate */
		unsigned nsac              : 8;	/* data read access-time-2 in CLK cycles (NSAC*100) */
		unsigned taac              : 8;	/* data read access-time-1 */
		unsigned                   : 6;
		unsigned csd_structure     : 2;	/* CSD structure */

		/* bits 64 - 95 */
		unsigned c_size            : 6;	/* Device size (bit 16-21) */
		unsigned                   : 6;
		unsigned dsr_imp           : 1;	/* DSR implemented */
		unsigned read_blk_misalign : 1;	/* Read block misalignment */
		unsigned write_blk_misalign: 1;	/* Write block misalignment */
		unsigned read_blk_partial  : 1;	/* Partial blocks for read allowed */
		unsigned read_blk_len      : 4;	/* Max. read data block length */
		unsigned ccc               : 12;/* Card command classes */

		/* bits 32 - 63 */
		unsigned wp_grp_size       : 7;	/* Write protect group size */
		unsigned sector_size       : 7;	/* Erase sector size */
		unsigned erase_blk_en      : 1;	/* Erase single block enable */
		unsigned                   : 1;
		unsigned c_size01          : 16;/* Device size (bit 0-15) */

		/* bits 0 - 31 */
		unsigned                   : 1;	/* not used, always 1 */
		unsigned crc               : 7;	/* 7-bit CRC */
		unsigned                   : 2;
		unsigned file_format       : 2;	/* File format */
		unsigned tmp_wp            : 1;	/* Temporary write protection */
		unsigned perm_wp           : 1;	/* Permanent write protection */
		unsigned copy              : 1;	/* Copy flag (OTP) */
		unsigned file_format_grp   : 1;	/* File format group */
		unsigned                   : 5;
		unsigned write_blk_partial : 1;	/* Partial blocks for write allowed */
		unsigned write_blk_len     : 4;	/* Max. write data block length */
		unsigned r2w_factor        : 3;	/* Write speed factor */
		unsigned                   : 2;
		unsigned wp_grp_enable     : 1;	/* Write protect group enable */
	} sd_v2;

	/* word layout */
	unsigned int u[4];
} mmc_sd_csd_t;



/*---------------------------------------------------------------------------
 *  Computation of CRC-7 0x48 polynom (x**7 + x**3 + 1)
 *  Used in MMC commands and responses.
 *---------------------------------------------------------------------------*/
static unsigned char calc_Crc7(unsigned char *pc, unsigned int len)
{
	unsigned int  i;
	unsigned int  ibit;
	unsigned char c;
	unsigned char crc = 0;

	for (i = 0; i < len; i++, pc++)
	{
		c = *pc;
		for (ibit = 0; ibit < 8; ibit++)
		{
			crc <<= 1;
			if ((c ^ crc) & 0x80)
				crc ^= 0x09;

			c <<= 1;
		}
		crc &= 0x7F;
	}

	return (crc << 1) | 0x01;
}

static euint8 sd_SendAppCmd(hwInterface *iface, euint8 cmd, euint32 arg, euint32 rca)
{
	euint8 resp;

	/* send CMD55 for application specific command */
	sd_Command(iface, MMC_CMD_APP_CMD, rca << 16);
	resp = sd_Resp8b(iface);
	if (0 == (resp & MMC_R1_ILL_CMD))
	{
		/* execute application specific command */
		sd_Command(iface, cmd, arg);
		resp = sd_Resp8b(iface);
	}

	return resp;
}

esint8 sd_Init(hwInterface *iface)
{
	int i;
	euint8 resp;
	euint32 opcondarg = 0;
	euint32 respCheck = 0;
	euint32 respOCR   = 0;

	iface->cardRCA  = 1;
	iface->cardSDHC = 0;

	/* Try to send reset command up to 100 times */
	i = 1000;
	do
	{
		sd_Command(iface, MMC_CMD_GO_IDLE_STATE, 0);
		resp = sd_Resp8b(iface);
	} while (resp != MMC_BC_RST_ACK && i--);

	if (resp != MMC_BC_RST_ACK)
	{
		if (resp == 0xff)
		{
			return -1;
		}
		else
		{
			sd_Resp8bError(iface, resp);
			return -2;
		}
	}


	/* check for High Capacity card */
	sd_Command(iface, SDCARD_CMD_SEND_IF_COND, SD_CHECK_PATTERN);
	resp = sd_Resp8b(iface);
	if (0 == (resp & MMC_R1_ILL_CMD))
	{
		/* check for valid response */
		respCheck = sd_Resp32b(iface);
		if (SD_CHECK_PATTERN == respCheck)
		{
			opcondarg |= SD_HIGH_CAPACITY;
		}
		else
		{
			/* this must be something strange --> not supported */
			return -3;
		}
	}

	/*
		First try to initialise a SD card; it must be handled differently:
		send application specific command 41 (ignored by MMC cards)
	 */
	i = 1024;
	do
	{
		/* send ACMD41 'SDCARD_CMD_APP_OP_COND' */
		resp = sd_SendAppCmd(iface, SDCARD_CMD_APP_OP_COND, opcondarg, 0);
		if (MMC_BC_CMD_ACK == resp)
		{
			/* get OCR info CMD58 */
			sd_Command(iface, MMC_CMD_READ_OCR, 0);
			if (MMC_BC_CMD_ACK == sd_Resp8b(iface))
			{
				respOCR = sd_Resp32b(iface);
				if ((respOCR & SD_HIGH_CAPACITY) != 0)
				{
					/* card supports block addressing mode */
					iface->cardSDHC = 1;
				}
			}
			/* SD card detected */
			return 0;
		}
	} while (0 == (resp & MMC_R1_ILL_CMD) && i--);

	/* Wait till card is ready initialising (returns 0 on CMD1) */
	/* Try up to 32000 times. */

	/* check for MMC card */
	i = 32000;
	do
	{
		sd_Command(iface, MMC_CMD_SEND_OP_COND, 0);
		resp = sd_Resp8b(iface);
		if (resp != MMC_BC_CMD_ACK)
			sd_Resp8bError(iface, resp);
	} while (resp == MMC_BC_RST_ACK && i--);

	if (resp != MMC_BC_CMD_ACK)
	{
		sd_Resp8bError(iface, resp);
		return -3;
	}

	return 0;
}
/*****************************************************************************/

void sd_Command(hwInterface *iface, euint8 cmd, euint32 param)
{
	euint8 frame[6];
	int i;

	frame[0] = cmd;
	frame[1] = (euint8)(param >> 24);
	frame[2] = (euint8)(param >> 16);
	frame[3] = (euint8)(param >> 8);
	frame[4] = (euint8)(param);
	frame[5] = calc_Crc7(frame, 5);;

	if_spiSend(iface, 0xff);

	for (i = 0; i < 6; ++i)
	{
		if_spiSend(iface, frame[i]);
	}

	if_spiSend(iface, 0xff);	/* eat empty command - response */
}
/*****************************************************************************/

euint8 sd_Resp8b(hwInterface *iface)
{
	int i;
	euint8 resp;

	/* Response will come after 1 - 8 pings */
	for (i = 0; i < 8; ++i)
	{
		resp = if_spiSend(iface, 0xff);
		if (resp != 0xff)
			break;
	}

	return resp;
}
/*****************************************************************************/

euint16 sd_Resp16b(hwInterface *iface)
{
	euint16 resp;

	resp  = (sd_Resp8b(iface) << 8) & 0xff00;
	resp |= if_spiSend(iface, 0xff);

	return resp;
}
/*****************************************************************************/

euint32 sd_Resp32b(hwInterface *iface)
{
	union
	{
		euint8	c[4];
		euint32	u;
	} res;

	/* bytes are reversed --> little endian word as result */
	res.c[3] = if_spiSend(iface, 0xff);
	res.c[2] = if_spiSend(iface, 0xff);
	res.c[1] = if_spiSend(iface, 0xff);
	res.c[0] = if_spiSend(iface, 0xff);

	return res.u;
}
/*****************************************************************************/

void sd_Resp8bError(hwInterface *iface, euint8 value)
{
	NOT_USED(iface);

	switch (value)
	{
		case 0x40 :
			DBG((TXT("Argument out of bounds.\n")));
			break;
		case 0x20 :
			DBG((TXT("Address out of bounds.\n")));
			break;
		case 0x10 :
			DBG((TXT("Error during erase sequence.\n")));
			break;
		case 0x08 :
			DBG((TXT("CRC failed.\n")));
			break;
		case 0x04 :
			DBG((TXT("Illegal command.\n")));
			break;
		case 0x02 :
			DBG((TXT("Erase reset (see SanDisk docs p5-13).\n")));
			break;
		case 0x01 :
			DBG((TXT("Card is initialising.\n")));
			break;
		default :
			DBG((TXT("Unknown error 0x%x (see SanDisk docs p5-13).\n"), value));
			break;
	}
}
/*****************************************************************************/

esint8 sd_State(hwInterface *iface)
{
	eint16 value;

	sd_Command(iface, MMC_CMD_SEND_STATUS, 0);
	value = sd_Resp16b(iface);

	switch (value)
	{
		case 0x000 :
			return 1;
			break;
		case 0x0001 :
			DBG((TXT("Card is Locked.\n")));
			break;
		case 0x0002 :
			DBG((TXT("WP Erase Skip, Lock/Unlock Cmd Failed.\n")));
			break;
		case 0x0004 :
			DBG((TXT("General / Unknown error -- card broken?.\n")));
			break;
		case 0x0008 :
			DBG((TXT("Internal card controller error.\n")));
			break;
		case 0x0010 :
			DBG((TXT("Card internal ECC was applied, but failed to correct the data.\n")));
			break;
		case 0x0020 :
			DBG((TXT("Write protect violation.\n")));
			break;
		case 0x0040 :
			DBG((TXT("An invalid selection, sectors for erase.\n")));
			break;
		case 0x0080 :
			DBG((TXT("Out of Range, CSD_Overwrite.\n")));
			break;
		default :
			if (value > 0x00FF)
			{
				sd_Resp8bError(iface, (euint8)(value >> 8));
			}
			else
			{
				DBG((TXT("Unknown error: 0x%x (see SanDisk docs p5-14).\n"), value));
			}
			break;
	}
	return -1;
}
/*****************************************************************************/

/* ****************************************************************************
 * WAIT ?? -- FIXME
 * CMDWRITE
 * WAIT
 * CARD RESP
 * WAIT
 * DATA BLOCK OUT
 *      START BLOCK
 *      DATA
 *      CHKS (2B)
 * BUSY...
 */

esint8 sd_writeSector(hwInterface *iface, euint32 address, euint8 *buf)
{
	euint32 place;
	int i;
	int t = 0;

	if (iface->cardSDHC == 0)
	{
		place = 512 * address;
	}
	else
	{
		place = address;
	}
	sd_Command(iface, MMC_CMD_WRITE_BLOCK, place);

	sd_Resp8b(iface);			/* Card response */

	if_spiSend(iface, MMC_BC_START);	/* Start block */
	for (i = 0; i < 512; ++i)
	{
		if_spiSend(iface, buf[i]);	/* Send data */
	}
	if_spiSend(iface, 0xff);	/* Checksum part 1 */
	if_spiSend(iface, 0xff);	/* Checksum part 2 */

	if_spiSend(iface, 0xff);

	while (if_spiSend(iface, 0xff) != 0xff)
	{
		t++;
	}

	return 0;
}
/*****************************************************************************/

/* ****************************************************************************
 * WAIT ?? -- FIXME
 * CMDREAD
 * WAIT
 * CARD RESP
 * WAIT
 * DATA BLOCK IN
 * 		START BLOCK
 * 		DATA
 * 		CHKS (2B)
 */

esint8 sd_readSector(hwInterface *iface, euint32 address, euint8 *buf, euint16 len)
{
	euint8 cardresp;
	euint8 firstblock;
	euint8 c;
	int fb_timeout = 0xffff;
	int i;
	euint32 place;

	if (iface->cardSDHC == 0)
	{
		place = 512 * address;
	}
	else
	{
		place = address;
	}
	sd_Command(iface, MMC_CMD_READ_SINGLE_BLOCK, place);

	cardresp = sd_Resp8b(iface);	/* Card response */

	/* Wait for startblock */
	do
	{
		firstblock = sd_Resp8b(iface);
	} while (firstblock == 0xff && fb_timeout--);

	if (cardresp != MMC_BC_CMD_ACK || firstblock != MMC_BC_START)
	{
		sd_Resp8bError(iface, firstblock);
		return -1;
	}

	for (i = 0; i < 512; ++i)
	{
		c = if_spiSend(iface, 0xff);
		if (i < len)
			buf[i] = c;
	}

	/* Checksum (2 byte) - ignore for now */
	if_spiSend(iface, 0xff);
	if_spiSend(iface, 0xff);

	return 0;
}
/*****************************************************************************/

/* ****************************************************************************
 calculates size of card from CSD
 (extension by Martin Thomas, inspired by code from Holger Klabunde)
 (reworked by Willi Thei, support for High Capacity SD cards)
 */
esint8 sd_getDriveSize(hwInterface *iface, euint32 *drive_size)
{
	euint8 cardresp;
	euint8 iob[16];
	int i, j;
	mmc_sd_csd_t *pCSD;
	unsigned int ds;	/* This will only work upto 4 GiB cards !! */

	/* read Card Specific Data */
	sd_Command(iface, MMC_CMD_SEND_CSD, 0);

	do
	{
		cardresp = sd_Resp8b(iface);
	} while (cardresp != MMC_BC_START);

	DBG((TXT("CSD:")));
	/* do the byte revers for each word on read */
	for (i = 0; i < 16; i += 4)
	{
		for (j = i + 3; j >= i; --j)
		{
			iob[j] = if_spiSend(iface, 0xff);
			DBG((TXT(" %02x"), iob[j]));
		}
	}
	DBG((TXT("\n")));

	/* Checksum (2 byte) - ignore for now */
	if_spiSend(iface, 0xff);
	if_spiSend(iface, 0xff);

	/* extract size info from CSD structure */
	pCSD = (mmc_sd_csd_t *)&iob[0];
	/* check for High Capacity SD card */
	if (1 == pCSD->sd.csd_structure)
	{
		unsigned int c_size = (pCSD->sd_v2.c_size << 16) + pCSD->sd_v2.c_size01;
		ds = (c_size + 1) << (9 + 10);
	}
	else
	{
		unsigned int c_size, c_size_mult, read_blk_len;

		read_blk_len = pCSD->sd.read_blk_len;
		c_size_mult  = pCSD->sd.c_size_mult;
		c_size = (pCSD->sd.c_size << 2) + pCSD->sd.c_size01;
		ds = (c_size + 1) * (1 << (c_size_mult + 2)) * (1 << read_blk_len);
	}
	*drive_size = (euint32)ds;

	return 0;
}
