/*====================================================================
* Project:  examples
* Function: phyCORE-TC39xB test for using Embedded FileSystem Library
*
* Copyright HighTec EDV-Systeme GmbH 1982-2019
*====================================================================*/

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#ifdef __TC161__
#include "system_tc2x.h"
#endif /* __TC161__ */
#ifdef __TC162__
#include "system_tc3x.h"
#endif /* __TC162__ */
#include "timer.h"
#include "led.h"
#include "uart_poll.h"

#include "efs.h"
#include "efs_ls.h"
#include "efs_mkfs.h"
#include "efs_ui.h"
#include "efs_time.h"
#include "interfaces/efsl_dbg_printf.h"

#define rprintf			efsl_debug_printf

#ifndef BAUDRATE
#define BAUDRATE		115200
#endif /* BAUDRATE */

#define SYSTIME_CLOCK	1000	/* timer event rate [Hz] */

#define SECTOR_SIZE		512


static EmbeddedFileSystem efs;
static EmbeddedFile filer, filew;
static DirList list;
#ifdef DATE_TIME_SUPPORT
static efsl_real_time_t efsTime;
#endif /* DATE_TIME_SUPPORT */
static euint32 e;
static unsigned char buf[SECTOR_SIZE+1];

static char LogFileName[] = "logTC_5.txt";

/* counter for time services */
static volatile unsigned int timeInTicks = 0;

/* timing control */
void TimerCallback(void)
{
	++timeInTicks;
}

static __inline unsigned int systime_get(void)
{
	return timeInTicks;
}

static void uart_puts(char *s)
{
	while (*s)
		_out_uart((unsigned char)*s++);
}

static int uart_putc(int ch)
{
	_out_uart((unsigned char)ch);
	return ch;
}

static void benchmark(void)
{
	signed char res;
	char bmfile[] = "bm2.txt";
	unsigned int starttime, deltat, bytes;
	int error;
	unsigned short l = 100;

	if ((res = efs_init(&efs, 0)) != 0)
	{
		rprintf("efs_init failed with %i\n", res);
		return;
	}

	rmfile(&efs.myFs, (euint8 *)bmfile);

	if (file_fopen(&filew, &efs.myFs, bmfile, 'w') != 0)
	{
		rprintf("\nfile_open for %s failed", bmfile);
		fs_umount(&efs.myFs);
		return;
	}

	rprintf("Write benchmark start - write to file %s (%i bytes/write)\n", bmfile, l);

	bytes = 0;
	error = 0;
	starttime = systime_get();	/* millisec. */

	do
	{
		if (file_write(&filew, l, buf) != l)
		{
			error = 1;
		}
		else
		{
			bytes += l;
		}
		deltat = systime_get() - starttime;
	} while ((deltat < 5000U) && !error);

	if (error)
	{
		rprintf("An error occured during write! [error=%d,%d bytes written]\n",
				ioman_getError(((IOManager *)(&efs.myIOman))), bytes);
		fs_umount(&efs.myFs);
		return;
	}

	file_fclose(&filew);
	fs_flushFs(&(efs.myFs));	/* close & flushing included in time */

	deltat = systime_get() - starttime;
	rprintf("%u bytes written in %u ms (%u KBytes/sec)\n",
			bytes, deltat, (((bytes / deltat) * SYSTIME_CLOCK) / 1024));

	rprintf("Read benchmark start - from file %s (in %i bytes blocks)\n", bmfile, l);

	if (file_fopen(&filer, &efs.myFs, bmfile, 'r') != 0)
	{
		rprintf("\nfile_open for %s failed", bmfile);
		fs_umount(&efs.myFs);
		return;
	}

	bytes = 0;
	error = 0;
	starttime = systime_get();	/* millisec. */

	while ((e = file_read(&filer, l, buf)) != 0)
	{
		bytes += (unsigned int)e;
	}

	file_fclose(&filer);

	deltat = systime_get() - starttime;
	rprintf("%u bytes read in %u ms (%u KBytes/sec)\n",
			bytes, deltat, (((bytes / deltat) * SYSTIME_CLOCK) / 1024));

	fs_umount(&efs.myFs);
}

static void EFSLDemo(int readOnly)
{
	signed char res;

	LEDON(0);

	rprintf("CARD init...");

	if ((res = efs_init(&efs, 0)) != 0)
	{
		rprintf("failed with %i\n", res);
	}
	else
	{
		rprintf("ok\n");

		LEDOFF(0);

		rprintf("\nDirectory of 'root':\n");
		ls_openDir(&list, &(efs.myFs), "/");
		while (ls_getNext(&list) == 0)
		{
			list.currentEntry.FileName[LIST_MAXLENFILENAME-1] = '\0';
#ifdef DATE_TIME_SUPPORT
			fs_unpackTimeDate(&efsTime, list.currentEntry.FileTime, list.currentEntry.FileDate);
			rprintf("%s (%li bytes) [%04u-%02u-%02u %02u:%02u:%02u]\n",
					list.currentEntry.FileName, list.currentEntry.FileSize,
					(unsigned)efsTime.year, (unsigned)efsTime.month, (unsigned)efsTime.day,
					(unsigned)efsTime.hours, (unsigned)efsTime.minutes,
					(unsigned)efsTime.seconds);
#else
			rprintf("%s ( %li bytes )\n", list.currentEntry.FileName,
										  list.currentEntry.FileSize);
#endif /* DATE_TIME_SUPPORT */
		}

		LEDON(0);

		if (file_fopen(&filer, &efs.myFs, LogFileName, 'r') == 0)
		{
			rprintf("\nFile %s open. Content:\n", LogFileName);
			while ((e = file_read(&filer, SECTOR_SIZE, buf)) != 0)
			{
				buf[e] = '\0';
				uart_puts((char *)buf);
			}
			rprintf("\n");
			file_fclose(&filer);
		}

		LEDOFF(0);

		/* do write tests if permitted */
		if (!readOnly)
		{
			if (file_fopen(&filew, &efs.myFs, LogFileName, 'a') == 0)
			{
				rprintf("\nFile %s open for append. Appending...", LogFileName);
				strcpy((char *)buf, "Martin hat's angehaengt\r\n");
				if (file_write(&filew, strlen((char *)buf), buf) == strlen((char *)buf))
				{
					rprintf("ok\n");
				}
				else
				{
					rprintf("failed\n", LogFileName);
				}
				file_fclose(&filew);
			}

			LEDON(0);

			if (file_fopen(&filer, &efs.myFs, LogFileName, 'r') == 0)
			{
				rprintf("\nFile %s open. Content:\n", LogFileName);
				while ((e = file_read(&filer, SECTOR_SIZE, buf)) != 0)
				{
					buf[e] = '\0';
					uart_puts((char *)buf);
				}
				rprintf("\n");
				file_fclose(&filer);
			}

			LEDOFF(0);
		}

		fs_umount(&efs.myFs);
	}
}

static void showMenu(void)
{
	uart_puts("\r\nSelect your action\r\n");
	uart_puts("  B: start EFSL read/write benchmark\r\n");
	uart_puts("  D: demo of EFSL functionality (read + write)\r\n");
	uart_puts("  R: demo of EFSL functionality (read only)\r\n");
	uart_puts("  T: display system time\r\n\n");
}

int main(void)
{
	unsigned char c;
	int flag = 0;

	InitLED();

	SYSTEM_Init();

	/* enable Code and Data caches */
	SYSTEM_EnaDisCache(1);

	/* initialise timer at SYSTIME_CLOCK rate */
	TimerInit(SYSTIME_CLOCK);
	TimerSetHandler(TimerCallback);

	_init_uart(BAUDRATE);

#ifdef __TC161__
	uart_puts("\n\n\rTriCore Filesystem-Demo (P:AURIX L:efsl)\r\n");
#endif /* __TC161__ */
#ifdef __TC162__
	uart_puts("\n\n\rTriCore Filesystem-Demo (P:AURIX 2G L:efsl)\r\n");
#endif /* __TC162__ */
	uart_puts("efsl AT91-Interface and this Demo-Application\r\n");
	uart_puts("done by Martin Thomas, Kaiserslautern, Germany\r\n");
	uart_puts("efsl TriCore-Interface done by Willi Thei, HighTec, Germany\r\n\n");

	/* init efsl debug-output */
	efsl_debug_devopen(uart_putc);

	showMenu();

	for (;;)
	{
		if (_poll_uart(&c))
		{
			switch (c)
			{
				case 'B' :
					benchmark();
					break;
				case 'D' :
					EFSLDemo(0);
					break;
				case 'R' :
					EFSLDemo(1);
					break;
				case 'T' :
					{
						unsigned int h, m, s, ms;
						unsigned int t = systime_get();
						s  = t / SYSTIME_CLOCK;
						ms = t % SYSTIME_CLOCK;
						m  = s / 60;
						s %=  60;
						h  = m / 60;
						m %= 60;
						rprintf("\nSystem Time: %02d:%02d:%02d.%03d\n", h, m, s, ms);
					}
					break;
				default :
					rprintf("\nYou pressed the \"%c\" key\n", (char)c);
					break;
			}

			/* toggle LED */
			if (flag)
			{
				flag = 0;
				LEDOFF(0);
			}
			else
			{
				flag = 1;
				LEDON(0);
			}

			showMenu();
		}
	}

	return EXIT_SUCCESS;
}
