/*****************************************************************************\
*              efs - General purpose Embedded Filesystem library              *
*          ---------------------------------------------------------          *
*                                                                             *
* Filename : tricore_spi.c                                                    *
* Description : This file contains the functions needed to use efs for        *
*               accessing files on an SD-card connected to a phyCORE-TC39xB   *
*               or to a phyCORE-TC29xB board                                  *
*                                                                             *
* This library is free software; you can redistribute it and/or               *
* modify it under the terms of the GNU Lesser General Public                  *
* License as published by the Free Software Foundation; either                *
* version 2.1 of the License, or (at your option) any later version.          *
*                                                                             *
* This library is distributed in the hope that it will be useful,             *
* but WITHOUT ANY WARRANTY; without even the implied warranty of              *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU           *
* Lesser General Public License for more details.                             *
*                                                                             *
*                                                    (c)2019 Willi Thei      *
*                                                                             *
\*****************************************************************************/


/*****************************************************************************/
#include <machine/wdtcon.h>

#include "tc_inc_path.h"

#include TC_INCLUDE(TCPATH/IfxPort_reg.h)
#include TC_INCLUDE(TCPATH/IfxQspi_reg.h)
#include TC_INCLUDE(TCPATH/IfxQspi_bf.h)

#include "interfaces/sd_spi.h"

/* macro for avoiding warnings about unused function arguments */
#define NOT_USED(x)		(void)(x)


/*****************************************************************************/

#define MMC_SLOW_CLOCKRATE	400000
#define MMC_FAST_CLOCKRATE	12500000


/*--------------- SD card interface hardware definitions -----------------*/

/*
	SD card interface is connected to QSPI5 interface

	Port	Function	Type	Note
	P15.10	MRST5A		I		Master Receive Slave Transmit
	P15.11	SLSO52		O		/CS_SD Chip Select
	P15.12	/CD_SD		I		card detect signal
	P15.14	MTSR5		O		Master Transmit Slave Receive
	P15.15	SCLK5		O		SCLK_SD
 */

/* Macros for accessing hardware related stuff */
#define CS_SDCARD			2		/* SLSO52 */
#define CS_SDCARD_ECON		QSPI5_ECON2
#define CS_SDCARD_SLSOMSK	(1u << (CS_SDCARD + IFX_QSPI_SSOC_OEN_OFF))

/* Macros for accessing MMC Chip Select signal */
/* phyCORE-TC39xB|phyCORE-TC29xB: P15.11 */
#define MMC_CS_PIN			P11
#define MMC_CS_DISABLE()	(P15_OUT.B.MMC_CS_PIN = 1)
#define MMC_CS_ENABLE()		(P15_OUT.B.MMC_CS_PIN = 0)
#define MMC_CS_MANUAL()		(P15_IOCR8.B.PC11 = 0x10)
#define MMC_CS_AUTO()		(P15_IOCR8.B.PC11 = 0x13)


void if_spiInit(hwInterface *iface)
{
	int i;

	/* prevent multiple base initialisation */
	if (0 != QSPI5_CLC.U)
	{
		unlock_wdtcon();				/* remove ENDINIT protection */
		QSPI5_CLC.U = 0x0;				/* activate module */
		(void)QSPI5_CLC.U;				/* read back to get effective */
		P15_PDR1.U = 0x00000800;		/* fast speed (all pins), MRST5 at TTL level */
		lock_wdtcon();					/* re-enable ENDINIT protection */

		/* configure port pins */
		MMC_CS_MANUAL();				/* SLSOx manual */
		MMC_CS_DISABLE();				/* not selected */
		P15_IOCR12.B.PC14 = 0x13;		/* MTSR5 */
		P15_IOCR12.B.PC15 = 0x13;		/* SCLK5 */

		/* program QSPI parameters */
#ifdef __TC161__
		QSPI5_GLOBALCON.U	= (15 << IFX_QSPI_GLOBALCON_EXPECT_OFF)
							| (1 << IFX_QSPI_GLOBALCON_SI_OFF);
#endif /* __TC161__ */
#ifdef __TC162__
		QSPI5_GLOBALCON.U	= (1 << IFX_QSPI_GLOBALCON_RESETS_OFF)
							| (1 << IFX_QSPI_GLOBALCON_CLKSEL_OFF)	/* CLKSEL=fPER */
							| (15 << IFX_QSPI_GLOBALCON_EXPECT_OFF)
							| (1 << IFX_QSPI_GLOBALCON_SI_OFF);
#endif /* __TC162__ */
		QSPI5_GLOBALCON1.U	= (4 << IFX_QSPI_GLOBALCON1_PT1_OFF)
							| (1 << IFX_QSPI_GLOBALCON1_USREN_OFF)
							| (1 << IFX_QSPI_GLOBALCON1_PT2EN_OFF)
							| (1 << IFX_QSPI_GLOBALCON1_PT1EN_OFF)
							| (1 << IFX_QSPI_GLOBALCON1_RXEN_OFF)
							| (1 << IFX_QSPI_GLOBALCON1_TXEN_OFF)
							| (0x7F << IFX_QSPI_GLOBALCON1_ERRORENS_OFF);
		/* enable SLSOx, low active */
		QSPI5_SSOC.U = CS_SDCARD_SLSOMSK;
	}
	else
	{
		/* disable QSPI5 */
		QSPI5_GLOBALCON.B.EN = 0;

		/* activate manual chip select */
		MMC_CS_MANUAL();
		/* ... and deselect */
		MMC_CS_DISABLE();
	}

	/* slow during init */
	if_spiSetSpeed(0xFF);

	/* enable QSPI5 */
	QSPI5_GLOBALCON.B.EN = 1;

	/* Send 21 spi commands with card not selected */
	for (i = 0; i < 21; ++i)
		if_spiSend(iface, 0xFF);

	/* activate automatic chip select */
	MMC_CS_AUTO();

	/* send at least 80 cycles 1s */
	for (i = 0; i < 10; ++i)
		if_spiSend(iface, 0xFF);
}
/*****************************************************************************/

void if_spiSetSpeed(euint8 speed)
{
	unsigned int baud = (speed == 0 ? MMC_FAST_CLOCKRATE : MMC_SLOW_CLOCKRATE);

	/* reprogram ECONx for SLSOx */
	if (baud == MMC_FAST_CLOCKRATE)
	{
		/* fast speed */
		/* 12.5 MHz ==> 80ns SCLK: Q=4,A=2,B=1,C=1 */
		CS_SDCARD_ECON.U = (1 << IFX_QSPI_ECON_C_OFF)
						 | (1 << IFX_QSPI_ECON_B_OFF)
						 | ((2 - 1) << IFX_QSPI_ECON_A_OFF)
						 | ((4 - 1) << IFX_QSPI_ECON_Q_OFF);
	}
	else
	{
		/* slow speed */
		/* 400 kHz ==> 2500ns SCLK: Q=50,A=4,B=3,C=3 */
		CS_SDCARD_ECON.U = (3 << IFX_QSPI_ECON_C_OFF)
						 | (3 << IFX_QSPI_ECON_B_OFF)
						 | (( 4 - 1) << IFX_QSPI_ECON_A_OFF)
						 | ((50 - 1) << IFX_QSPI_ECON_Q_OFF);
	}
}
/*****************************************************************************/

euint8 if_spiSend(hwInterface *iface, euint8 outgoing)
{
	unsigned int data;

	NOT_USED(iface);

	/* we need 2 free entries */
	while (QSPI5_STATUS.B.TXFIFOLEVEL > (4 - 2))
		;
	/* we need 8 bit for data transfer */
	QSPI5_BACONENTRY.U	= (CS_SDCARD << IFX_QSPI_BACON_CS_OFF)
						| (IFX_QSPI_BACON_MSB_MSK << IFX_QSPI_BACON_MSB_OFF)
						| ((8-1) << IFX_QSPI_BACON_DL_OFF)
						| (IFX_QSPI_BACON_LAST_MSK << IFX_QSPI_BACON_LAST_OFF);
	/* write data */
	QSPI5_DATAENTRY0.U	= outgoing;

	/* wait until all data received */
	while (QSPI5_STATUS.B.RXFIFOLEVEL != 2)
		;

	/* ... and read the data */
	data = QSPI5_RXEXIT.U;
	(void)QSPI5_RXEXIT.U;

	return (euint8)data;
}
