/*====================================================================
* Project:  Board Support Package
* Function: test program for ethernet on TC3x
*
* Copyright HighTec EDV-Systeme GmbH 1982-2020
*====================================================================*/

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "serialio.h"

#include "system_tc3x.h"
#if defined(USE_ETH_MODULE)
#include "eth_tc3x.h"
#else
#include "geth_tc3x.h"
#endif /* USE_ETH_MODULE */

#ifndef BAUDRATE
#define BAUDRATE	115200
#endif /* BAUDRATE */

#define ETHMAC		"\x0\x50\xC2\x3B\xAA\xA6"
#if 1
#define MY_IP		"\xC0\xA8\x0\xC7"			/* 192.168.0.199 */
#else
#define MY_IP		"\xC0\xA8\x16\xC7"			/* 192.168.22.199 */
#endif
#define ETHMAC_BC	"\xFF\xFF\xFF\xFF\xFF\xFF"

/* definition of constants: no need of ntohs() for a little endian target */
#define FRAME_IP		0x0008
#define FRAME_ARP		0x0608
#define OP_ARP_REQ		0x0100
#define OP_ARP_ANS		0x0200
#define HARDW_ETH10		0x0100	/* hardware-type 10Mbps Ethernet */
#define IP_HLEN_PLEN	0x0406	/* MAC = 6 byte long, IP = 4 byte long */
#define IP_VER_IHL		0x0045	/* IPv4, Header Length = 5x32 bit */
#define PROT_ICMP		0x0100	/* internet Control Message Protocol */
#define ICMP_ECHO		0x0008	/* message is an echo request */
#define ICMP_ECHO_REPLY	0x0000	/* message is an echo reply */

#define MAXBUFSIZE	2048

static unsigned char rxbuf[MAXBUFSIZE];
static unsigned char txbuf[MAXBUFSIZE];

static unsigned int rxcnt = 0;
static unsigned int rxerr = 0;
static unsigned int txcnt = 0;
static unsigned int txerr = 0;
static unsigned int rx_strange_cnt = 0;


static __inline unsigned short ntohs(unsigned short Data)
{
	return (Data >> 8) | (Data << 8);
}

static unsigned short CalcChecksum(void *Start, unsigned int Count)
{
	unsigned long Sum = 0;
	unsigned short *piStart = (unsigned short *)Start;

	while (Count > 1)							/* sum words */
	{
		Sum += *piStart++;
		Count -= 2;
	}

	if (Count)									/* add left-over byte, if any */
		Sum += *(unsigned char *)piStart;

	while (Sum >> 16)							/* fold 32-bit sum to 16 bits */
		Sum = (Sum & 0xFFFF) + (Sum >> 16);

	return (unsigned short)~Sum;
}

static int processFrame(void)
{
	int tosend = 0;

	/* test for own MAC or broadcast */
	if (memcmp(ETHMAC, rxbuf, 6))
	{
		if (memcmp(ETHMAC_BC, rxbuf, 6))
		{
			++rx_strange_cnt;
		}
	}

	if ((*(unsigned short *)(rxbuf+12) == FRAME_ARP)
	    && (*(unsigned short *)(rxbuf+20) == OP_ARP_REQ)
	    && !memcmp(MY_IP, rxbuf+38, 4))
	{
		/* reply to an ARP request */
		/* ethernet header */
		memcpy(txbuf, rxbuf+6, 6);
		memcpy(txbuf+6, ETHMAC, 6);
		*(unsigned short *)(txbuf+12) = FRAME_ARP;

		/* ARP entry */
		*(unsigned short *)(txbuf+14) = HARDW_ETH10;
		*(unsigned short *)(txbuf+16) = FRAME_IP;
		*(unsigned short *)(txbuf+18) = IP_HLEN_PLEN;
		*(unsigned short *)(txbuf+20) = OP_ARP_ANS;
		memcpy(txbuf+22, ETHMAC, 6);
		memcpy(txbuf+28, MY_IP, 4);
		memcpy(txbuf+32, rxbuf+6, 6);
		memcpy(txbuf+38, rxbuf+28, 4);

		/* send ARP answer */
		tosend = 42;
	}
	else
	if ((*(unsigned short *)(rxbuf+12) == FRAME_IP)
	    && ((*(unsigned short *)(rxbuf+14) & 0x00FF) == IP_VER_IHL)
	    && ((*(unsigned short *)(rxbuf+22) & 0xFF00) == PROT_ICMP)
	    && !memcmp(MY_IP, rxbuf+30, 4)
	    && ((*(unsigned short *)(rxbuf+34) & 0x00FF) == ICMP_ECHO))
	{
		unsigned short dataCnt = ntohs(*(unsigned short *)(rxbuf+16));

		/* reply to an ICMP_ECHO request */
		/* ethernet header */
		memcpy(txbuf, rxbuf+6, 6);
		memcpy(txbuf+6, ETHMAC, 6);
		*(unsigned short *)(txbuf+12) = FRAME_IP;

		/* IP header */
		*(unsigned short *)(txbuf+14) = IP_VER_IHL;
		*(unsigned short *)(txbuf+16) = ntohs(dataCnt);
		*(unsigned short *)(txbuf+18) = 0;
		*(unsigned short *)(txbuf+20) = 0;
		*(unsigned short *)(txbuf+22) = 0x0140;	/* protocol + TTL (time to live) */
		*(unsigned short *)(txbuf+24) = 0;		/* head checksum */
		memcpy(txbuf+26, MY_IP, 4);
		memcpy(txbuf+30, rxbuf+26, 4);
		*(unsigned short *)(txbuf+24) = CalcChecksum(txbuf+14, 20);	/* head checksum */

		/* ICMP field */
		*(unsigned short *)(txbuf+34) = ICMP_ECHO_REPLY;
		*(unsigned short *)(txbuf+36) = 0;		/* ICMP message checksum */
		memcpy(txbuf+38, rxbuf+38, dataCnt-24);
		*(unsigned short *)(txbuf+36) = CalcChecksum(txbuf+34, dataCnt-20);	/* ICMP message checksum */

		/* send ICMP echo reply */
		tosend = (int)dataCnt + 14;
	}
	if (tosend)
	{
		if (eth_tc3x_send((const char *)txbuf, tosend) >= tosend)
		{
			++txcnt;
		}
		else
		{
			++txerr;
		}
	}

	return 0;
}

int main(void)
{
	int err;

	SYSTEM_Init();

#if 1 /* WT 2020-05-18 */
	serialio_init(BAUDRATE);
	printf("\rTestLAN_GETH_Phytec started\r\n");
#endif /* WT 2020-05-18 */

	err = eth_tc3x_init((const unsigned char *)ETHMAC);

	while (!err)
	{
		if (eth_tc3x_poll())
		{
			err = eth_tc3x_read((char *)rxbuf, sizeof(rxbuf));
			if (err > 0)
			{
				++rxcnt;
				err = processFrame();
#if 1 /* WT 2020-05-18 */
				printf("rxcnt=%d; rxerr=%d; txcnt=%d; txerr=%d\r\n", rxcnt, rxerr, txcnt, txerr);
#endif /* WT 2020-05-18 */
			}
			else
			{
				++rxerr;
				err = 0;
			}
		}
	}

	return 0;
}
