/*====================================================================
* Project:  Board Support Package
* Function: Standalone driver for GETH gigabit ethernet module of TC3x
*
* Copyright HighTec EDV-Systeme GmbH 1982-2020
*====================================================================*/

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <machine/wdtcon.h>

#include "tc_inc_path.h"

#include TC_INCLUDE(TCPATH/IfxPort_reg.h)
#include TC_INCLUDE(TCPATH/IfxGeth_reg.h)
#include TC_INCLUDE(TCPATH/IfxGeth_bf.h)
#include TC_INCLUDE(TCPATH/IfxScu_reg.h)
#include TC_INCLUDE(TCPATH/IfxScu_bf.h)
#include TC_INCLUDE(TCPATH/IfxStm_reg.h)
#include TC_INCLUDE(TCPATH/IfxStm_bf.h)

#include "system_tc3x.h"
#include "geth_tc3x.h"


#ifndef USE_RMII_ONLY
#define USE_RMII_ONLY	0
#endif /* USE_RMII_ONLY */

#ifndef ETHERMTU
#define ETHERMTU	1518
#endif /* ETHERMTU */

/* private data structure of driver */
typedef struct
{
	unsigned char	etheraddr[6];	/* own Ethernet Address     */
	unsigned char	phy_addr;		/* own PHYs address         */
	unsigned char	phy_ready;		/* ready state PHY device   */
	union
	{
		struct
		{
			unsigned short	phy_id2;	/* recognized PHY ID 2 */
			unsigned short	phy_id1;	/* recognized PHY ID 1 */
		} s;
		unsigned int phy_id;			/* combined ID value */
	} u;
} em_Data_t;

static em_Data_t emData;

static Ifx_GETH * const eth  = (Ifx_GETH *)&MODULE_GETH;

/* ========================================================================= */

/*****************************************/
/* Memory allocation for Ethernet module */
/*****************************************/
/*
	We use the 2nd 16 kiB block of DSPR_CPU1 for Ethernet buffers and descriptors.
 */

#define ETH_SRAM_BASE		0x60000000

#define ETHRAM_START		(ETH_SRAM_BASE + 16 * 1024)
#define ETHRAM_SIZE			(16 * 1024)

#define ETH_BUFFER_SIZE		((ETHERMTU + 0xFF) & 0xFF00)
#define ETH_MAX_BUFFER		(ETHRAM_SIZE / ETH_BUFFER_SIZE / 2)


/* RX descriptor */
typedef struct _eth_rxdesc_t
{
	/* RDES0: Buffer1 address */
	union
	{
		/* normal descriptor (read format) */
		unsigned char	*buffer1;
		/* normal descriptor (write back format) */
		struct
		{
			unsigned int	OVT		: 16;	/* Outer VLAN Tag */
			unsigned int	IVT		: 16;	/* Inner VLAN Tag */
		} wb;
		/* context descriptor */
		unsigned int	timeStampLow;
		unsigned int	u;
	} rdes0;

	/* RDES1: reserved (in read format) */
	/* normal descriptor (write back format) [only valid for last descriptor] */
	union
	{
		struct
		{
			unsigned int	PT		:  3;	/* Payload Type */
			unsigned int	IPHE	:  1;	/* IP Header Error */
			unsigned int	IPV4	:  1;	/* IPv4 header present */
			unsigned int	IPV6	:  1;	/* IPv6 header present */
			unsigned int	IPCB	:  1;	/* IP Checksum Bypassed */
			unsigned int	IPCE	:  1;	/* IP Payload Error */
			unsigned int	PMT		:  4;	/* PTP Message Type */
			unsigned int	PFT		:  1;	/* PTP Frame Type */
			unsigned int	PV		:  1;	/* PTP Version */
			unsigned int	TSA		:  1;	/* TimeStamp Available */
			unsigned int	TD		:  1;	/* Timestamp Dropped */
			unsigned int	OPC		: 16;	/* OAM Sub-Type Code or MAC Control Packet opcode */
		} wb;
		/* context descriptor */
		unsigned int	timeStampHigh;
		unsigned int	u;
	} rdes1;

	/* RDES2: Buffer2 address */
	union
	{
		/* normal descriptor (read format) */
		unsigned char	*buffer2;
		/* normal descriptor (write back format) */
		struct
		{
			unsigned int	HL		: 10;	/* L3/L4 Header Length */
			unsigned int	ARPNR	: 1;	/* ARP Reply Not Generated */
			unsigned int			: 3;	/* reserved */
			unsigned int	ITS		: 1;	/* Inner VLAN Tag filter Status */
			unsigned int	OTS		: 1;	/* Outer VLAN Tag filte Status */
			unsigned int	SAF		: 1;	/* Source Address Filter Fail */
			unsigned int	DAF		: 1;	/* Destination Address Filter Fail */
			unsigned int	HF		: 1;	/* Hash Filter Status */
			unsigned int	MADRM	: 8;	/* MAC Address Match or Hash Value */
			unsigned int			: 5;	/* reserved */
		} wb;
		unsigned int	u;
	} rdes2;

	/* RDES3 */
	union
	{
		/* normal descriptor (read format) */
		struct
		{
			unsigned int				: 24;	/* reserved */
			unsigned int	BUF1V		:  1;	/* Buffer 1 Address Valid */
			unsigned int	BUF2V		:  1;	/* Buffer 2 Address Valid */
			unsigned int				:  4;	/* reserved */
			unsigned int	IOC			:  1;	/* Interrupt On Completion enabled */
			unsigned int	OWN			:  1;	/* Own bit */
		} rd;
		/* normal descriptor (write back format) */
		struct
		{
			unsigned int	PL			: 15;	/* Packet Length */
			unsigned int	ES			: 1;	/* Error Summary */
			unsigned int	LT			: 3;	/* Length/Type Field */
			unsigned int	DE			: 1;	/* Dribble Bit Error */
			unsigned int	RE			: 1;	/* Receive Error */
			unsigned int	OE			: 1;	/* Overflow Error */
			unsigned int	RWT			: 1;	/* Receive Watchdog Timeout */
			unsigned int	GP			: 1;	/* Giant Packet */
			unsigned int	CE			: 1;	/* CRC Error */
			unsigned int	RS0V		: 1;	/* Receive Status RDES0 Valid */
			unsigned int	RS1V		: 1;	/* Receive Status RDES1 Valid */
			unsigned int	RS2V		: 1;	/* Receive Status RDES2 Valid */
			unsigned int	LD			: 1;	/* Last Descriptor */
			unsigned int	FD			: 1;	/* First Descriptor */
			unsigned int	CTXT		: 1;	/* Context Type (0==normal,1==context) */
			unsigned int	OWN			: 1;	/* Own bit */
		} wb;
		/* context descriptor */
		struct
		{
			unsigned int				: 29;	/* reserved */
			unsigned int	DE			: 1;	/* Descriptor Error */
			unsigned int	CTXT		: 1;	/* Context Type (0==normal,1==context) */
			unsigned int	OWN			: 1;	/* Own bit */
		} cd;
		unsigned int	u;
	} rdes3;
} eth_rxdesc_t;


/* TX descriptor */
typedef struct _eth_txdesc_t
{
	/* TDES0: Buffer1 address OR TimeStampLow */
	union
	{
		/* normal descriptor (read format) */
		unsigned char	*buffer1;
		/* normal descriptor (write back format) */
		/* context descriptor */
		unsigned int	timeStampLow;
	} tdes0;

	/* TDES1: Buffer2 address OR Buffer1 (high part of) address OR TimeStampHigh */
	union
	{
		/* normal descriptor (read format) */
		unsigned char	*buffer2;
		unsigned char	*buffer1_high;
		/* normal descriptor (write back format) */
		/* context descriptor */
		unsigned int	timeStampHigh;
	} tdes1;

	/* TDES2 */
	union
	{
		/* normal descriptor (read format) */
		struct
		{
			unsigned int	TBS1		: 14;	/* Transmit Buffer 1 Size */
			unsigned int	VTIR		:  2;	/* VLAN Tag Insertion or Replacement */
			unsigned int	TBS2		: 14;	/* Transmit Buffer 2 Size */
			unsigned int	TTSE		:  1;	/* Transmit TimeStamp Enable */
			unsigned int	IOC			:  1;	/* Interrupt On Completion */
		} rd;
		/* context descriptor */
		struct
		{
			unsigned int	MSS			: 14;	/* Maximum Segment Size */
			unsigned int				: 2;	/* reserved */
			unsigned int	IVT			: 16;	/* Inner VLAN Tag */
		} cd;
		unsigned int	u;
	} tdes2;

	/* TDES3 */
	union
	{
		/* normal descriptor (read format) */
		struct
		{
			unsigned int	FL_TPL		: 15;	/* Packet Length or TCP Payload Length */
			unsigned int	TPL			: 1;	/* TCP Payload Length (if TSE==1) */
			unsigned int	CIC			: 2;	/* Checksum Insertion Control */
			unsigned int				: 1;	/* reserved */
			unsigned int	THL			: 4;	/* TCP Header Length (if TSE==1) */
			unsigned int	SAIC		: 3;	/* SA Insertion Control */
			unsigned int	CPC			: 2;	/* CRC Pad Control */
			unsigned int	LD			: 1;	/* Last Descriptor */
			unsigned int	FD			: 1;	/* First Descriptor */
			unsigned int	CTXT		: 1;	/* Context Type (0==normal,1==context) */
			unsigned int	OWN			: 1;	/* Own bit */
		} rd;
		/* normal descriptor (write back format) */
		struct
		{
			unsigned int	IHE			: 1;	/* IP Header Error */
			unsigned int	DB			: 1;	/* Deferred Bit */
			unsigned int	UF			: 1;	/* Underflow error */
			unsigned int	ED			: 1;	/* Excessive Deferral */
			unsigned int	CC			: 4;	/* Collision Count */
			unsigned int	EC			: 1;	/* Excessive Collision */
			unsigned int	LC			: 1;	/* Late Collision */
			unsigned int	NC			: 1;	/* No Carrier */
			unsigned int	LOC			: 1;	/* Loss Of Carrier */
			unsigned int	PCE			: 1;	/* Payload Checksum Error */
			unsigned int	FF			: 1;	/* Frame Flushed */
			unsigned int	JT			: 1;	/* Jabber Timeout */
			unsigned int	ES			: 1;	/* Error Summary */
			unsigned int				: 1;	/* reserved */
			unsigned int	TTSS		: 1;	/* Tx TimeStamp Status */
			unsigned int				: 5;	/* reserved */
			unsigned int	DE			: 1;	/* Descriptor Error */
			unsigned int				: 4;	/* reserved */
			unsigned int	LD			: 1;	/* Last Descriptor */
			unsigned int	FD			: 1;	/* First Descriptor */
			unsigned int	CTXT		: 1;	/* Context Type (0==normal,1==context) */
			unsigned int	OWN			: 1;	/* Own bit */
		} wb;
		/* context descriptor */
		struct
		{
			unsigned int	VT			: 16;	/* VLAN Tag */
			unsigned int	VLTV		: 1;	/* VLAN Tag Valid */
			unsigned int	IVLTV		: 1;	/* Inner VLAN Tag Valid */
			unsigned int	IVTIR		: 2;	/* Inner VLAN Tag Insert or Replace */
			unsigned int				: 3;	/* reserved */
			unsigned int	DE			: 1;	/* Descriptor Error */
			unsigned int				: 2;	/* reserved */
			unsigned int	TCMSSV		: 1;	/* One-Step Timestamp Correction Input or MSS Valid */
			unsigned int	OSTC		: 1;	/* One-Step Timestamp Correction Enable */
			unsigned int				: 2;	/* reserved */
			unsigned int	CTXT		: 1;	/* Context Type (0==normal) */
			unsigned int	OWN			: 1;	/* Own bit */
		} cd;
		unsigned int	u;
	} tdes3;
} eth_txdesc_t;


/* memory allocation structure for ethernet RAM */
typedef struct _eth_memory_t
{
	eth_rxdesc_t  rxdesc[ETH_MAX_BUFFER];	/* RX descriptors */

	eth_txdesc_t  txdesc[ETH_MAX_BUFFER];	/* TX descriptors */

	unsigned char rxdata[ETH_MAX_BUFFER][ETH_BUFFER_SIZE];	/* buffer for RX */
	unsigned char txdata[ETH_MAX_BUFFER][ETH_BUFFER_SIZE];	/* buffer for TX */

	eth_rxdesc_t *currRx;					/* current RX descriptor to process */
	eth_txdesc_t *currTx;					/* current TX descriptor to process (MAC) */
	eth_txdesc_t *headTx;					/* current TX descriptor to process (application) */
	eth_rxdesc_t *lastRx;					/* last RX descriptor (end of ring mark) */
	eth_txdesc_t *lastTx;					/* last TX descriptor (end of ring mark) */
} eth_memory_t;

/* pointer to ethernet RAM */
static eth_memory_t * const eth_Memory = (eth_memory_t *)ETHRAM_START;


/* ========================================================================= */

/*****************************************/
/* Pin configuration for Ethernet module */
/*****************************************/

/*
	Pin usage of Gigabit Ethernet module (in RMII mode)

				TriBoard-TC3x5	TriBoard-TC3x4

	RMII_TXD1		P11.2		P11.2
	RMII_TXD0		P11.3		P11.3
	RMII_TXEN		P11.6		P11.6
	RMII_MDC		P21.2 (O5)	P02.8 (O6)
	RMII_MDIOD		P21.3
	RMII_MDIOA					P00.0

	RMII_RXD1		P11.9		P11.9
	RMII_RXD0		P11.10		P11.10
	RMII_CRSDV		P11.11		P11.11
	RMII_REFCLK		P11.12		P11.12

	PHY_INTB		[P15.8]		[P10.3]


	Pin usage of Gigabit Ethernet module
	[remark: all boards with V2.0 or higher are RGMII capable]

				AppKit-TC3x7	TriBoard-TC3x7	TriBoard-TC3x6
								TriBoard-TC3x9

	RGMII_TXD3		P11.0			P11.0			P11.0
	RGMII_TXD2		P11.1			P11.1			P11.1
	RGMII_TXD1		P11.2			P11.2			P11.2
	RGMII_TXD0		P11.3			P11.3			P11.3
	RGMII_TXCLK		P11.4			P11.4			P11.4
	RGMII_TXCTL		P11.6			P11.6			P11.6
	RGMII_MDC		P12.0			P12.0			P12.0
	RGMII_MDIOC		P12.1			P12.1			P12.1

	RGMII_RXD3A		P11.7			P11.7			P11.7
	RGMII_RXD2A		P11.8			P11.8			P11.8
	RGMII_RXD1A		P11.9			P11.9			P11.9
	RGMII_RXD0A		P11.10			P11.10			P11.10
	RGMII_RXCTL		P11.11			P11.11			P11.11
	RGMII_RXCLK		P11.12			P11.12			P11.12

	RGMII_GREFCLK	P11.5			P11.5			P11.5
	PHY_INTB		[P10.8|P02.1]	[P15.8]			[P10.3]


	Remarks:
	GETH_GPCTL: can only be written in SVC mode
	Pn_PDRm (m=0,1), Pn_ESR, Pn_PDISC, Pn_LPCRm (m=0-7): ENDINIT protected
	Pn_PCSR, Pn_ACCENm (m=1,2): SAFETY ENDINIT protected
 */

#define PAD_OUT_PP_ALT7		0x17
#define PAD_OUT_PP_ALT6		0x16
#define PAD_OUT_PP_ALT5		0x15



static void eth_init_gpio(void)
{
	unlock_wdtcon();
	/* set IO speed */
#if (USE_RMII_ONLY > 0)
	P11_PDR0.U = 0x23223322;	/* pins 2,3,6: Speed Grade 4 for RMII driver */
	P11_PDR1.U = 0x22233332;	/* pins 9..12: Speed Grade 4 for RMII driver */
# if 0 /* TriBoard_TC3x4 */
	P02_PDR1.U = 0x00000003;	/* pin 8: Speed Grade 4 for RMII driver */
# else
	P21_PDR0.U = 0x00002322;	/* pin 2: Speed Grade 4 for RMII driver */
# endif /* TriBoard_TC3x4 */
#else
	P11_PDR0.U = 0x33333333;	/* pins 0..7:  Speed Grade 4 for RGMII driver */
	P11_PDR1.U = 0x22233333;	/* pins 8..12: Speed Grade 4 for RGMII driver */
	P12_PDR0.U = 0x00002223;	/* pin 0: Speed Grade 4 for RGMII driver */
#endif /* USE_RMII_ONLY > 0 */
	lock_wdtcon();

	unlock_safety_wdtcon();
#if (USE_RMII_ONLY > 0)
	/* select ethernet outputs via fast RMII */
	P11_PCSR.U = 0x0000004C;	/* P2,P3,P6 */
#else
	/* select ethernet outputs via fast RGMII */
	P11_PCSR.U = 0x0000005F;	/* P0,P1,P2,P3,P4,P6 */
#endif /* USE_RMII_ONLY > 0 */
	lock_safety_wdtcon();

#if (USE_RMII_ONLY > 0)
	/* select alternate outputs */
# if 0 /* TriBoard_TC3x4 */
	P02_IOCR8.B.PC8 = PAD_OUT_PP_ALT6;	/* MDC */
	/* MDIO: controlled by hardware ==> no settings */
# else
	P21_IOCR0.B.PC2 = PAD_OUT_PP_ALT5;	/* MDC */
	/* MDIO: controlled by hardware ==> no settings */
# endif /* TriBoard_TC3x4 */
	P11_IOCR4.B.PC6 = PAD_OUT_PP_ALT6;	/* TXEN */
	P11_IOCR0.B.PC3 = PAD_OUT_PP_ALT6;	/* TXD0 */
	P11_IOCR0.B.PC2 = PAD_OUT_PP_ALT6;	/* TXD1 */
#else
	/* select alternate outputs */
	P12_IOCR0.B.PC0 = PAD_OUT_PP_ALT6;	/* MDC */
	/* MDIO: controlled by hardware ==> no settings */
	P11_IOCR0.U = 0xB0B0B0B0;	/* TXD0, TXD1, TXD2, TXD3 */
#if 0
	P11_IOCR4.U = 0x00B000B8;
	P11_IOCR8.U = 0x00000000;
	P11_IOCR12.B.PC12 = 0x00;
#else
	/* remark: do not touch inputs */
	P11_IOCR4.B.PC6 = PAD_OUT_PP_ALT6;	/* TXCTL */
	P11_IOCR4.B.PC4 = PAD_OUT_PP_ALT7;	/* TXCLK */
#endif
#endif /* USE_RMII_ONLY > 0 */

	/* select alternate inputs */
#if (USE_RMII_ONLY > 0)
# if 0 /* TriBoard_TC3x4 */
	eth->GPCTL.U = 0;			/* ALTI0==0 (MDIOA from P00.0), all other 0 */
# else
	eth->GPCTL.U = 3;			/* ALTI0==3 (MDIOD from P21.3), all other 0 */
# endif /* TriBoard_TC3x4 */
#else
#if 0
	/* kept for documentation */
	eth->GPCTL.B.ALTI0 = 2;		/* MDIOC from P12.1 */
	eth->GPCTL.B.ALTI1 = 0;		/* REFCLK from P11.12 */
	eth->GPCTL.B.ALTI4 = 0;		/* CRSDV from P11.11 */
	eth->GPCTL.B.ALTI6 = 0;		/* RXD0 from P11.10 */
	eth->GPCTL.B.ALTI7 = 0;		/* RXD1 from P11.9 */
#else
	eth->GPCTL.U = 2;			/* ALTI0==2 (MDIOC from P12.1 ), all other 0 */
#endif
#endif /* USE_RMII_ONLY > 0 */

#if (USE_RMII_ONLY > 0)
	/* select RMII */
	eth->GPCTL.B.EPR = 4;
#else
	/* select RGMII */
	eth->GPCTL.B.EPR = 1;
#endif /* USE_RMII_ONLY > 0 */
}


/**************************/
/* support for PHY device */
/**************************/

/* PHY internal register */
#define ETH_PHY_CTRL			0			/* PHY Control Register */
#define ETH_PHY_STAT			1			/* PHY Status Register */
#define ETH_PHY_ID1				2			/* PHY Identifier 1 */
#define ETH_PHY_ID2				3			/* PHY Identifier 2 */
#define ETH_PHY_AUTO_AD			4			/* PHY Auto-Negotiation Advertisement Register */
#define ETH_PHY_AUTO_CAP		5			/* PHY Auto-Negotiation Capabilities Register */

/* LANTIQ PEF7071 specific registers */
#define ETH_PHY_GCTRL			9			/* Gigabit Control Register */
#define ETH_PHY_GSTAT			10			/* Gigabit Status Register */
#define ETH_PHY_MIICTRL			23			/* Media-Independent Interface Control */
#define ETH_PHY_MIISTAT			24			/* Media-Independent Interface Status */

/* RTL8201FI specific registers */
#define ETH_PHY_PAGSR			31			/* Page Selection Register */
/* page 7 regs */
#define ETH_PHY_RMSR			16			/* RMII Mode Selection Register */
#define ETH_PHY_IWLFR			19			/* Interrupt, WOL Enable, and LEDs Function Register */

/* RTL8211F specific registers */
#define ETH_PHY_GBCR			9			/* 1000Base-T Control Register */
#define ETH_PHY_GBSR			10			/* 1000Base-T Status Register */
#define ETH_PHY_GBESR			15			/* 1000Base-T Extended Status Register */
#define ETH_PHY_PHYSR			26			/* PHY Specific Status Register */
#define ETH_PHY_EPAGSR			31			/* Extension Page Selection Register */
/* extension page 0xD04 regs */
#define ETH_PHY_LCR				16			/* LED Control Register */
#define ETH_PHY_EEELCR			17			/* EEE LED Control Register */

/* TI DP83867IRRGZ specific registers */
#define ETH_PHY_CFG1			9			/* 1000BASE-T Configuration Register (CFG1) */
#define ETH_PHY_STS1			10			/* Status Register 1 (STS1) */
#define ETH_PHY_REGCR			13			/* Register Control Register (REGCR) */
#define ETH_PHY_ADDAR			14			/* Address or Data Register (ADDAR) */
#define ETH_PHY_1KSCR			15			/* 1000BASE-T Status Register (1KSCR) */
#define ETH_PHY_PHYSTS			17			/* PHY Status Register (PHYSTS) */
#define ETH_PHY_MICR			18			/* MII Interrupt Control Register (MICR) */
#define ETH_PHY_LEDCR1			24			/* LED Configuration Register 1 (LEDCR1) */
#define ETH_PHY_LEDCR2			25			/* LED Configuration Register 2 (LEDCR2) */
#define ETH_PHY_LEDCR3			26			/* LED Configuration Register 3 (LEDCR3) */
#define ETH_PHY_RGMIICTL		50			/* RGMII Control Register (RGMIICTL) */
#define ETH_PHY_STRAP_STS1		110			/* Strap Configuration Status Register 1 (STRAP_STS1) */
#define ETH_PHY_STRAP_STS2		111			/* Strap Configuration Status Register 2 (STRAP_STS2) */
#define ETH_PHY_RGMIIDCTL		134			/* RGMII Delay Control Register (RGMIIDCTL) */
#define ETH_PHY_SYNC_FIFO_CTRL	233			/* Sync FIFO Control (SYNC_FIFO_CTRL) */
#define ETH_PHY_IO_MUX_CFG		368			/* I/O Configuration (IO_MUX_CFG) */


/* PHY control bits */
#define ETH_PHY_CTRL_RST		(1 << 15)	/* 1=PHY Reset */
#define ETH_PHY_CTRL_LPBK		(1 << 14)	/* 1=PHY Loopback */
#define ETH_PHY_CTRL_SPEED		(1 << 13)	/* 1=100Mbps, 0=10Mbps */
#define ETH_PHY_CTRL_ANEG_EN	(1 << 12)	/* 1=Enable Auto negotiation */
#define ETH_PHY_CTRL_PDN		(1 << 11)	/* 1=PHY Power Down mode */
#define ETH_PHY_CTRL_MII_DIS	(1 << 10)	/* 1=MII 4 bit interface disabled */
#define ETH_PHY_CTRL_ANEG_RST	(1 << 9)	/* 1=Reset Auto negotiate */
#define ETH_PHY_CTRL_DPLX		(1 << 8)	/* 1=Full Duplex, 0=Half Duplex */
#define ETH_PHY_CTRL_COLTST		(1 << 7)	/* 1=MII Collision Test */
/* RTL8211F additional bits */
#define ETH_PHY_CTRL_SPEED1		(1 << 6)	/* 1=1000Mbps (if bit 13 == 0) */
#define ETH_PHY_CTRL_UNIEN		(1 << 5)	/* 1=Uni-Directional Enable (without link) */

/* PHY status bits */
#define ETH_PHY_STAT_CAP_T4		(1 << 15)	/* 1=100Base-T4 capable */
#define ETH_PHY_STAT_CAP_TXF	(1 << 14)	/* 1=100Base-X full duplex capable */
#define ETH_PHY_STAT_CAP_TXH	(1 << 13)	/* 1=100Base-X half duplex capable */
#define ETH_PHY_STAT_CAP_TF		(1 << 12)	/* 1=10Mbps full duplex capable */
#define ETH_PHY_STAT_CAP_TH		(1 << 11)	/* 1=10Mbps half duplex capable */
#define ETH_PHY_STAT_CAP_SUPR	(1 << 6)	/* 1=recv mgmt frames with not preamble */
#define ETH_PHY_STAT_ANEG_ACK	(1 << 5)	/* 1=ANEG has completed */
#define ETH_PHY_STAT_REM_FLT	(1 << 4)	/* 1=Remote Fault detected */
#define ETH_PHY_STAT_CAP_ANEG	(1 << 3)	/* 1=Auto negotiate capable */
#define ETH_PHY_STAT_LINK		(1 << 2)	/* 1=valid link */
#define ETH_PHY_STAT_JAB		(1 << 1)	/* 1=10Mbps jabber condition */
#define ETH_PHY_STAT_EXREG		(1 << 0)	/* 1=extended registers implemented */

/* PHY Auto-Negotiation Advertisement bits */
#define ETH_PHY_AUTO_AD_NP		(1 << 15)	/* 1=PHY requests exchange of Next Page */
#define ETH_PHY_AUTO_AD_ACK		(1 << 14)	/* 1=got link code word from remote */
#define ETH_PHY_AUTO_AD_RF		(1 << 13)	/* 1=advertise remote fault */
#define ETH_PHY_AUTO_AD_T4		(1 << 9)	/* 1=PHY is capable of 100Base-T4 */
#define ETH_PHY_AUTO_AD_TX_FDX	(1 << 8)	/* 1=PHY is capable of 100Base-TX FDPLX */
#define ETH_PHY_AUTO_AD_TX_HDX	(1 << 7)	/* 1=PHY is capable of 100Base-TX HDPLX */
#define ETH_PHY_AUTO_AD_10_FDX	(1 << 6)	/* 1=PHY is capable of 10Base-T FDPLX */
#define ETH_PHY_AUTO_AD_10_HDX	(1 << 5)	/* 1=PHY is capable of 10Base-T HDPLX */
#define ETH_PHY_AUTO_AD_CSMA	(1 << 0)	/* 1=PHY is capable of 802.3 CMSA */

/* RTL8211F specific registers: PHY Specific Status Register */
#define ETH_PHY_PHYSR_SPEED		(3 << 4)
#define ETH_PHY_PHYSR_SP10M		(0 << 4)	/*   10 Mbps */
#define ETH_PHY_PHYSR_SP100M	(1 << 4)	/*  100 Mbps */
#define ETH_PHY_PHYSR_SP1000M	(2 << 4)	/* 1000 Mbps */
#define ETH_PHY_PHYSR_DUPLEX	(1 << 3)	/* full duplex */

/* LANTIQ PEF7071 specific register bits: Media-Independent Interface Status */
#define ETH_PHY_MIISTAT_SPEED	(3 << 0)	/* speed mask */
#define ETH_PHY_MIISTAT_SP10M	(0 << 0)	/*   10 Mbps */
#define ETH_PHY_MIISTAT_SP100M	(1 << 0)	/*  100 Mbps */
#define ETH_PHY_MIISTAT_SP1000M	(2 << 0)	/* 1000 Mbps */
#define ETH_PHY_MIISTAT_DPX		(1 << 3)	/* full duplex */


/* some PHY IDs (combined ID1 + ID2) */
#define PHY_ID_RTL8201F			0x001CC816		/* Realtek PHY RTL8201F */
#define PHY_ID_RTL8211F			0x001CC916		/* Realtek PHY RTL8211F */
#define PHY_ID_PEF7071			0xD565A401		/* LANTIQ PHY PEF7071 */
#define PHY_ID_DP83867IRRGZ		0x2000A231		/* TI DP83867IRRGZ */


static __inline unsigned short eth_read_phy(unsigned char phyaddr, unsigned char phyreg)
{
	unsigned int addrVal = eth->MAC_MDIO_ADDRESS.U;

	/* prepare MDIO address register value */
	addrVal &= (IFX_GETH_MAC_MDIO_ADDRESS_CR_MSK << IFX_GETH_MAC_MDIO_ADDRESS_CR_OFF);	/* keep clock settings */
	addrVal |= (((unsigned int)phyaddr & IFX_GETH_MAC_MDIO_ADDRESS_PA_MSK) << IFX_GETH_MAC_MDIO_ADDRESS_PA_OFF);
	addrVal |= (((unsigned int)phyreg  & IFX_GETH_MAC_MDIO_ADDRESS_RDA_MSK) << IFX_GETH_MAC_MDIO_ADDRESS_RDA_OFF);
	/* read: GOC_0=1,GOC_1=1 */
	addrVal |= (IFX_GETH_MAC_MDIO_ADDRESS_GOC_0_MSK << IFX_GETH_MAC_MDIO_ADDRESS_GOC_0_OFF);
	addrVal |= (IFX_GETH_MAC_MDIO_ADDRESS_GOC_1_MSK << IFX_GETH_MAC_MDIO_ADDRESS_GOC_1_OFF);
	addrVal |= (IFX_GETH_MAC_MDIO_ADDRESS_GB_MSK << IFX_GETH_MAC_MDIO_ADDRESS_GB_OFF);
	eth->MAC_MDIO_ADDRESS.U = addrVal;

	/* wait until PHY is ready */
	while (eth->MAC_MDIO_ADDRESS.B.GB)
		;

	return (unsigned short)eth->MAC_MDIO_DATA.B.GD;
}

static __inline void eth_write_phy(unsigned char phyaddr, unsigned char phyreg, unsigned short val)
{
	unsigned int addrVal = eth->MAC_MDIO_ADDRESS.U;

	/* prepare MDIO address register value */
	addrVal &= (IFX_GETH_MAC_MDIO_ADDRESS_CR_MSK << IFX_GETH_MAC_MDIO_ADDRESS_CR_OFF);	/* keep clock settings */
	addrVal |= (((unsigned int)phyaddr & IFX_GETH_MAC_MDIO_ADDRESS_PA_MSK) << IFX_GETH_MAC_MDIO_ADDRESS_PA_OFF);
	addrVal |= (((unsigned int)phyreg  & IFX_GETH_MAC_MDIO_ADDRESS_RDA_MSK) << IFX_GETH_MAC_MDIO_ADDRESS_RDA_OFF);
	/* write: GOC_0=1,GOC_1=0 */
	addrVal |= (IFX_GETH_MAC_MDIO_ADDRESS_GOC_0_MSK << IFX_GETH_MAC_MDIO_ADDRESS_GOC_0_OFF);
	addrVal |= (IFX_GETH_MAC_MDIO_ADDRESS_GB_MSK << IFX_GETH_MAC_MDIO_ADDRESS_GB_OFF);
	eth->MAC_MDIO_DATA.B.GD = val;
	eth->MAC_MDIO_ADDRESS.U = addrVal;

	/* wait until PHY is ready */
	while (eth->MAC_MDIO_ADDRESS.B.GB)
		;
}

static __inline int eth_check_phy(unsigned char phyaddr)
{
	unsigned short ctrl_data;
	unsigned short stat_data;

	ctrl_data = eth_read_phy(phyaddr, ETH_PHY_CTRL);
	stat_data = eth_read_phy(phyaddr, ETH_PHY_STAT);
	if ((ctrl_data == 0xFFFF) || ((ctrl_data == 0) && (stat_data == 0)))
		return -1;

	return 0;
}

static int eth_detect_phy(unsigned char *phyaddr)
{
	unsigned char addr;

	for (addr = 0; addr < 32; ++addr)
	{
		if (eth_check_phy(addr) == 0)
		{
			*phyaddr = addr;
			return 0;
		}
	}

	return -1;
}

/* use STM for timing */
#define TIME_SCALE_DN	100000
#define TIME_SCALE_UP	(1000000 / TIME_SCALE_DN)
/* shorter times do not work reliably ! */
#define PHY_TEST_TIME	100000		/* time interval for testing PHY state [us] (100 ms) */
#define PHY_MAX_WAIT	50			/* max wait count for end of auto-negotiation [5 s] */

#define HAL_DELAY_US(t)		hal_delay_us(t)

/* wait for <time> micro seconds */
/* beware of overflows: 100 us at (>=)43 MHz will overflow (if not scaled before multiplying) */
static void hal_delay_us(unsigned int time)
{
	unsigned int fSTM = (unsigned int)SYSTEM_GetStmClock();
	unsigned int stmWaitCount = (fSTM / TIME_SCALE_DN) * time / TIME_SCALE_UP;

	/* prepare compare register */
	STM0_CMP0.U = STM0_TIM0.U + stmWaitCount;
	STM0_CMCON.U = 31;
	/* Attention: keep this order, otherwise first match will trigger too soon */
	/* reset interrupt flag */
	STM0_ISCR.U = (IFX_STM_ISCR_CMP0IRR_MSK << IFX_STM_ISCR_CMP0IRR_OFF);
	/* enable compare match */
	STM0_ICR.B.CMP0EN = 1;
	/* wait for compare match */
	while (0 == STM0_ICR.B.CMP0IR)
		;
	STM0_ICR.B.CMP0EN = 0;
}

#if 1
unsigned short strap_sts1 = 0;
unsigned short strap_sts2 = 0;
unsigned short rgmiidctrl = 0;
unsigned short led_cr1 = 0;
unsigned short led_cr2 = 0;
unsigned short led_cr3 = 0;
#endif

static int eth_phy_autoneg(em_Data_t *data)
{
	unsigned short ctrl_val;
	unsigned short phy_stat;
	unsigned char  phyaddr = data->phy_addr;
	unsigned int   to_cnt;

	/* reset PHY */
	ctrl_val = ETH_PHY_CTRL_RST;
	eth_write_phy(phyaddr, ETH_PHY_CTRL, ctrl_val);
	/* initialise timeout object */
	to_cnt = PHY_MAX_WAIT;
	/* wait - with timeout - until RESET is finished */
	do
	{
		HAL_DELAY_US(PHY_TEST_TIME);
		ctrl_val = eth_read_phy(phyaddr, ETH_PHY_CTRL);
		if (!(ctrl_val & ETH_PHY_CTRL_RST))
		{
			break;
		}
		--to_cnt;
	} while (to_cnt);
	if (to_cnt == 0)
	{
		return 1;
	}

#if (USE_RMII_ONLY > 0)
	if (PHY_ID_RTL8201F == data->u.phy_id)
	{
		/* setup RTL8201FI PHY */
		/* select page 7 */
		ctrl_val = 7;
		eth_write_phy(phyaddr, ETH_PHY_PAGSR, ctrl_val);
		/* page 7 reg 16: activate RMII mode */
		ctrl_val = 0x0FF8;	/* RMII mode (10/100MBit) */
		eth_write_phy(phyaddr, ETH_PHY_RMSR, ctrl_val);
		/* page 7 reg 19: enable traditional LED functionality */
#if 0
		ctrl_val = 0x31;	/* LED1 == LINK100/ACT100; LED0 == LINK10/ACT10 */
#else
		ctrl_val = 0x01;	/* LED1 == LINK100; LED0 == ACTall */
#endif
		eth_write_phy(phyaddr, ETH_PHY_IWLFR, ctrl_val);
		ctrl_val = 0;		/* back to page 0 */
		eth_write_phy(phyaddr, ETH_PHY_PAGSR, ctrl_val);
	}
	else
	{
		/* ToDo: handle other PHYs used in RMII mode */
	}
#else
	if (PHY_ID_RTL8211F == data->u.phy_id)
	{
		/* setup RTL8211F PHY (taken from AppKit-TC387 SW) */
		/* select (undocumented) extension page 0xD08, reg 17 */
		/* according Linux sources this page/reg is used for controlling TX_DELAY (in RGMII) */
		ctrl_val = 0x0D08;
		eth_write_phy(phyaddr, ETH_PHY_EPAGSR, ctrl_val);
		ctrl_val = eth_read_phy(phyaddr, 17);
		ctrl_val |= 0x100;
		eth_write_phy(phyaddr, 17, ctrl_val);
		/* select extension page 0xD04, reg 16 = LED Control, reg 17 = EEE LED control */
		ctrl_val = 0x0D04;
		eth_write_phy(phyaddr, ETH_PHY_EPAGSR, ctrl_val);
#if 0
		ctrl_val = 0x8170;	/* remark: modification of reserved bits */
#else
		/* remark: reserved bit 15 used (no action without using it) ?? */
		ctrl_val = 0x8110;	/* LED1_LINK_1000 (bit 8), LED0_ACT (bit 4), [LED2 not connected] */
#endif
		eth_write_phy(phyaddr, ETH_PHY_LCR, ctrl_val);
		ctrl_val = 0;		/* disable EEE for all LEDs */
		eth_write_phy(phyaddr, ETH_PHY_EEELCR, ctrl_val);
		/* back to page 0 */
		eth_write_phy(phyaddr, ETH_PHY_EPAGSR, ctrl_val);
	}
	else if (PHY_ID_DP83867IRRGZ == data->u.phy_id)
	{
		/* setup SYNC_FIFO_CTRL (0xDF22) [see 8.3.2.1 (??)] */
		ctrl_val = (0 << 14) | 0x1F;	/* setup address */
		eth_write_phy(phyaddr, ETH_PHY_REGCR, ctrl_val);
		ctrl_val = ETH_PHY_SYNC_FIFO_CTRL;
		eth_write_phy(phyaddr, ETH_PHY_ADDAR, ctrl_val);
		ctrl_val = (1 << 14) | 0x1F;	/* setup data without post increment */
		eth_write_phy(phyaddr, ETH_PHY_REGCR, ctrl_val);
		ctrl_val = 0xDF22;
		eth_write_phy(phyaddr, ETH_PHY_ADDAR, ctrl_val);

		/* read some configuration regs */
#if 0
		ctrl_val = (0 << 14) | 0x1F;	/* setup address */
		eth_write_phy(phyaddr, ETH_PHY_REGCR, ctrl_val);
		ctrl_val = ETH_PHY_STRAP_STS1;
		eth_write_phy(phyaddr, ETH_PHY_ADDAR, ctrl_val);
		ctrl_val = (1 << 14) | 0x1F;	/* setup data without post increment */
		eth_write_phy(phyaddr, ETH_PHY_REGCR, ctrl_val);
		/* read strap configuration 1 */
		strap_sts1 = eth_read_phy(phyaddr, ETH_PHY_ADDAR);

		ctrl_val = (0 << 14) | 0x1F;	/* setup address */
		eth_write_phy(phyaddr, ETH_PHY_REGCR, ctrl_val);
		ctrl_val = ETH_PHY_STRAP_STS2;
		eth_write_phy(phyaddr, ETH_PHY_ADDAR, ctrl_val);
		ctrl_val = (1 << 14) | 0x1F;	/* setup data without post increment */
		eth_write_phy(phyaddr, ETH_PHY_REGCR, ctrl_val);
		/* read strap configuration 2 */
		strap_sts2 = eth_read_phy(phyaddr, ETH_PHY_ADDAR);
#else
		ctrl_val = (0 << 14) | 0x1F;	/* setup address */
		eth_write_phy(phyaddr, ETH_PHY_REGCR, ctrl_val);
		ctrl_val = ETH_PHY_STRAP_STS1;
		eth_write_phy(phyaddr, ETH_PHY_ADDAR, ctrl_val);
		ctrl_val = (2 << 14) | 0x1F;	/* setup data with post increment */
		eth_write_phy(phyaddr, ETH_PHY_REGCR, ctrl_val);
		/* read strap configuration 1 */
		strap_sts1 = eth_read_phy(phyaddr, ETH_PHY_ADDAR);
		/* read strap configuration 2 */
		strap_sts2 = eth_read_phy(phyaddr, ETH_PHY_ADDAR);
#endif
#if 1 /* WT 2020-05-18 */
		printf("PHY_STRAP_STS1 := 0x%04x; PHY_STRAP_STS2 := 0x%04x\r\n", strap_sts1, strap_sts2);
#endif /* WT 2020-05-18 */

#if 1 /* WT 2020-05-20 */
		ctrl_val = (0 << 14) | 0x1F;	/* setup address */
		eth_write_phy(phyaddr, ETH_PHY_REGCR, ctrl_val);
		ctrl_val = ETH_PHY_IO_MUX_CFG;
		eth_write_phy(phyaddr, ETH_PHY_ADDAR, ctrl_val);
		ctrl_val = (1 << 14) | 0x1F;	/* setup data without post increment */
		eth_write_phy(phyaddr, ETH_PHY_REGCR, ctrl_val);
		/* read I/O configuration */
		ctrl_val = eth_read_phy(phyaddr, ETH_PHY_ADDAR);
		printf("PHY_IO_MUX_CFG := 0x%04x\r\n", ctrl_val);
		/* change field CLK_O_SEL [12:8] to 0 ==> "00000: Channel A receive clock" */
		ctrl_val &= ~0x1F00;
#if 0 /* WT 2020-05-22 */
		/* change field CLK_O_SEL [12:8] to 0 ==> "01000: Channel A transmit clock" */
		ctrl_val |= 0x0800;
#endif /* WT 2020-05-22 */
		eth_write_phy(phyaddr, ETH_PHY_ADDAR, ctrl_val);
#endif /* WT 2020-05-20 */

		/* read LED configuration */
		led_cr1 = eth_read_phy(phyaddr, ETH_PHY_LEDCR1);
		led_cr2 = eth_read_phy(phyaddr, ETH_PHY_LEDCR2);
		led_cr3 = eth_read_phy(phyaddr, ETH_PHY_LEDCR3);
#if 0
		/* set "default" for LEDCR2 */
		eth_write_phy(phyaddr, ETH_PHY_LEDCR2, 0x4444);
#else
		/* invert LED_0 and LED_2 */
		eth_write_phy(phyaddr, ETH_PHY_LEDCR2, 0x4040);
#endif

#if 1 /* WT 2020-05-11 */
		ctrl_val = (0 << 14) | 0x1F;	/* setup address */
		eth_write_phy(phyaddr, ETH_PHY_REGCR, ctrl_val);
		ctrl_val = ETH_PHY_RGMIIDCTL;
		eth_write_phy(phyaddr, ETH_PHY_ADDAR, ctrl_val);
		ctrl_val = (1 << 14) | 0x1F;	/* setup data without post increment */
		eth_write_phy(phyaddr, ETH_PHY_REGCR, ctrl_val);
		/* set minimal delay for RX & TX */
//		ctrl_val = 0x11;
		ctrl_val = 0x77;
//		ctrl_val = 0x73;
//		ctrl_val = 0x71;
		eth_write_phy(phyaddr, ETH_PHY_ADDAR, ctrl_val);
#endif /* WT 2020-05-11 */

		ctrl_val = (0 << 14) | 0x1F;	/* setup address */
		eth_write_phy(phyaddr, ETH_PHY_REGCR, ctrl_val);
		ctrl_val = ETH_PHY_RGMIIDCTL;
		eth_write_phy(phyaddr, ETH_PHY_ADDAR, ctrl_val);
		ctrl_val = (1 << 14) | 0x1F;	/* setup data without post increment */
		eth_write_phy(phyaddr, ETH_PHY_REGCR, ctrl_val);
		/* read RGMII Delay Control configuration */
		rgmiidctrl = eth_read_phy(phyaddr, ETH_PHY_ADDAR);
#if 0 /* WT 2020-05-11 */
		ctrl_val = (0 << 14) | 0x1F;	/* setup address */
		eth_write_phy(phyaddr, ETH_PHY_REGCR, ctrl_val);
		ctrl_val = ETH_PHY_RGMIIDCTL;
		eth_write_phy(phyaddr, ETH_PHY_ADDAR, ctrl_val);
		ctrl_val = (1 << 14) | 0x1F;	/* setup data without post increment */
		eth_write_phy(phyaddr, ETH_PHY_REGCR, ctrl_val);
		/* set minimal delay for RX & TX */
		ctrl_val = 0x11;
		eth_write_phy(phyaddr, ETH_PHY_ADDAR, ctrl_val);
#endif /* WT 2020-05-11 */
	}
	else
	{
		/* ToDo: handle other PHYs used in RGMII mode */
	}
#endif /* USE_RMII_ONLY > 0 */

	phy_stat = eth_read_phy(phyaddr, ETH_PHY_STAT);
	/* check whether PHY is capable of doing auto-negotiation */
	if (phy_stat & ETH_PHY_STAT_CAP_ANEG)
	{
		/* set auto-negotiation advertisement capabilities */
		ctrl_val = ETH_PHY_AUTO_AD_TX_FDX
				 | ETH_PHY_AUTO_AD_TX_HDX
				 | ETH_PHY_AUTO_AD_10_FDX
				 | ETH_PHY_AUTO_AD_10_HDX
				 | ETH_PHY_AUTO_AD_CSMA;
		eth_write_phy(phyaddr, ETH_PHY_AUTO_AD, ctrl_val);

		/* start auto-negotiation */
		ctrl_val = ETH_PHY_CTRL_ANEG_EN
				 | ETH_PHY_CTRL_ANEG_RST;
	}
	else
	{
#if (USE_RMII_ONLY > 0)
		/* force 100Mbps full duplex */
		ctrl_val = ETH_PHY_CTRL_SPEED
				 | ETH_PHY_CTRL_DPLX;
#else
		/* force 1000Mbps full duplex */
		ctrl_val = ETH_PHY_CTRL_SPEED1
				 | ETH_PHY_CTRL_DPLX;
#endif /* USE_RMII_ONLY > 0 */
	}

	eth_write_phy(phyaddr, ETH_PHY_CTRL, ctrl_val);

	/* initialise timeout object */
	to_cnt = PHY_MAX_WAIT;

	/* wait - with timeout - until auto-negotiation is finished */
	do
	{
		HAL_DELAY_US(PHY_TEST_TIME);
		phy_stat = eth_read_phy(phyaddr, ETH_PHY_STAT);
#if 0
		if (phy_stat & ETH_PHY_STAT_ANEG_ACK)
#else
		if (phy_stat & ETH_PHY_STAT_LINK)
#endif
		{
			break;
		}
	} while (--to_cnt);

	/* check for result of autonegotiation */
	if (to_cnt)
	{
#if (USE_RMII_ONLY == 0)
		/* only usable for RGMII mode (ATM) */
		Ifx_GETH_MAC_CONFIGURATION mac_config = eth->MAC_CONFIGURATION;
#if 0 /* check different PHYs extended state */
		/* read extended PHY status */
		if (PHY_ID_RTL8211F == data->u.phy_id)
		{
			phy_stat = eth_read_phy(phyaddr, ETH_PHY_PHYSR);
			/* configure MAC according to selected speed + duplex */
			mac_config.B.DM = (phy_stat & ETH_PHY_PHYSR_DUPLEX) ? 1 : 0;
			switch (phy_stat & ETH_PHY_PHYSR_SPEED)
			{
				case ETH_PHY_PHYSR_SP1000M :
					mac_config.B.FES = 0;
					mac_config.B.PS  = 0;
					break;
				case ETH_PHY_PHYSR_SP100M :
					mac_config.B.FES = 1;
					mac_config.B.PS  = 1;
					break;
				case ETH_PHY_PHYSR_SP10M :
					mac_config.B.FES = 0;
					mac_config.B.PS  = 1;
					break;
			}
		}
		else if (PHY_ID_PEF7071 == data->u.phy_id)
		{
			phy_stat = eth_read_phy(phyaddr, ETH_PHY_MIISTAT);
			/* configure MAC according to selected speed + duplex */
			mac_config.B.DM = (phy_stat & ETH_PHY_MIISTAT_DPX) ? 1 : 0;
			switch (phy_stat & ETH_PHY_MIISTAT_SPEED)
			{
				case ETH_PHY_MIISTAT_SP1000M :
					mac_config.B.FES = 0;
					mac_config.B.PS  = 0;
					break;
				case ETH_PHY_MIISTAT_SP100M :
					mac_config.B.FES = 1;
					mac_config.B.PS  = 1;
					break;
				case ETH_PHY_MIISTAT_SP10M :
					mac_config.B.FES = 0;
					mac_config.B.PS  = 1;
					break;
			}
		}
		else
		{
			/* ToDo: how to cope with unknown PHYs */
		}
#else
		/* instead of handling individual PHYs ==> check MAC_PHYIF_CONTROL_STATUS */
		/* remark: this does only work in RGMII mode */
		mac_config.B.DM = eth->MAC_PHYIF_CONTROL_STATUS.B.LNKMOD;
		switch (eth->MAC_PHYIF_CONTROL_STATUS.B.LNKSPEED)
		{
			case 0 :	/*   2.5 MHz ==>   10 Mbps */
				mac_config.B.FES = 0;
				mac_config.B.PS  = 1;
				break;
			case 1 :	/*  25 MHz   ==>  100 Mbps */
				mac_config.B.FES = 1;
				mac_config.B.PS  = 1;
				break;
			case 2 :	/* 125 MHz   ==> 1000 Mbps */
				mac_config.B.FES = 0;
				mac_config.B.PS  = 0;
				break;
			default :
				break;
		}
		/* write back adapted MAC config */
		eth->MAC_CONFIGURATION = mac_config;
#endif /* check different PHYs extended state */
#else
		if (PHY_ID_RTL8201F == data->u.phy_id)
		{
			Ifx_GETH_MAC_CONFIGURATION mac_config = eth->MAC_CONFIGURATION;

			/* PHY_CTRL bits reflect result of autonegotiation */
			phy_stat = eth_read_phy(phyaddr, ETH_PHY_CTRL);
			/* configure MAC according to selected speed + duplex */
			mac_config.B.DM  = (phy_stat & ETH_PHY_CTRL_DPLX)  ? 1 : 0;
			mac_config.B.FES = (phy_stat & ETH_PHY_CTRL_SPEED) ? 1 : 0;

			/* write back adapted MAC config */
			eth->MAC_CONFIGURATION = mac_config;
		}
		else
		{
			/* ToDo: how to cope with unknown PHYs */
		}
#endif /* USE_RMII_ONLY == 0 */
	}

	return (to_cnt > 0 ? 0 : 1);
}

static int eth_phy_check_link(unsigned char phyaddr)
{
	return (eth_read_phy(phyaddr, ETH_PHY_STAT) & ETH_PHY_STAT_LINK) ? 0 : -1;
}

static int eth_init_phy(em_Data_t *data)
{
	int err = eth_detect_phy(&data->phy_addr);
	if (!err)
	{
		data->u.s.phy_id1 = eth_read_phy(data->phy_addr, ETH_PHY_ID1);
		data->u.s.phy_id2 = eth_read_phy(data->phy_addr, ETH_PHY_ID2);
#if 1 /* WT 2020-05-18 */
		printf("PHY [0x%08x] found at address %d\r\n", data->u.phy_id, data->phy_addr);
#endif /* WT 2020-05-18 */

		err = eth_phy_autoneg(data);
		if (!err)
		{
			err = eth_phy_check_link(data->phy_addr);
		}
	}

	return err;
}


/* initalise RX and TX data structures for DMA */
static void eth_init_memory(void)
{
	unsigned int i;

	/* clear all descriptor arrays */
	memset(eth_Memory->rxdesc, 0, sizeof(eth_Memory->rxdesc));
	memset(eth_Memory->txdesc, 0, sizeof(eth_Memory->txdesc));

	/* clear all data buffers */
	memset(eth_Memory->rxdata, 0, sizeof(eth_Memory->rxdata));
	memset(eth_Memory->txdata, 0, sizeof(eth_Memory->txdata));

	/* fill in descriptor list */
	for (i = 0; i < ETH_MAX_BUFFER; ++i)
	{
		eth_Memory->rxdesc[i].rdes0.buffer1 = &eth_Memory->rxdata[i][0];
		eth_Memory->rxdesc[i].rdes3.rd.BUF1V = 1;
		eth_Memory->rxdesc[i].rdes3.rd.IOC   = 1;
		eth_Memory->rxdesc[i].rdes3.rd.OWN   = 1;

		eth_Memory->txdesc[i].tdes0.buffer1 = &eth_Memory->txdata[i][0];
		eth_Memory->txdesc[i].tdes2.rd.TBS1 = ETH_BUFFER_SIZE;
	}

	eth_Memory->currRx = &eth_Memory->rxdesc[0];
	eth_Memory->currTx = &eth_Memory->txdesc[0];
	eth_Memory->headTx = &eth_Memory->txdesc[0];
	eth_Memory->lastRx = &eth_Memory->rxdesc[ETH_MAX_BUFFER - 1];
	eth_Memory->lastTx = &eth_Memory->txdesc[ETH_MAX_BUFFER - 1];

	/* configure descriptor info in DMA (using only channel 0) */
	eth->DMA_CH[0].RXDESC_LIST_ADDRESS.U = (unsigned int)&eth_Memory->rxdesc[0];
	eth->DMA_CH[0].RXDESC_TAIL_POINTER.U = (unsigned int)&eth_Memory->rxdesc[ETH_MAX_BUFFER - 1];
	eth->DMA_CH[0].RXDESC_RING_LENGTH.U  = ETH_MAX_BUFFER - 1;
	eth->DMA_CH[0].RX_CONTROL.U = (ETH_BUFFER_SIZE << IFX_GETH_DMA_CH_RX_CONTROL_RBSZ_X_0_OFF);
	eth->DMA_CH[0].TXDESC_LIST_ADDRESS.U = (unsigned int)&eth_Memory->txdesc[0];
	eth->DMA_CH[0].TXDESC_TAIL_POINTER.U = (unsigned int)&eth_Memory->txdesc[ETH_MAX_BUFFER - 1];
	eth->DMA_CH[0].TXDESC_RING_LENGTH.U  = ETH_MAX_BUFFER - 1;
}


/* ========================================================================= */

/* public interface for this module */

/* initialise ethernet hardware using given MAC address */
int eth_tc3x_init(const unsigned char *eth_mac)
{
	size_t i;

	/* check for GETH clock */
	/* ToDo: should this be part of PLL init (??) */
	if (0 == SCU_CCUCON5.B.GETHDIV)
	{
		unsigned int ccucon5 = SCU_CCUCON5.U;
		ccucon5 |= (1 << IFX_SCU_CCUCON5_GETHDIV_OFF);
		ccucon5 |= (1 << IFX_SCU_CCUCON5_UP_OFF);

		unlock_safety_wdtcon();
		SCU_CCUCON5.U = ccucon5;
		lock_safety_wdtcon();
	}

	/* This seems to be necessary for AppKit-TC397B (?) */
	/* do a kernel reset of GETH module */
	unlock_wdtcon();
	eth->KRST0.B.RST = 1;
	eth->KRST1.B.RST = 1;
	lock_wdtcon();
	/* wait until reset is finished */
	while (0 == eth->KRST0.B.RSTSTAT)
		;
	unlock_wdtcon();
	/* clear reset status */
	eth->KRSTCLR.B.CLR = 1;
	lock_wdtcon();

	/* Enable ETH module */
	unlock_wdtcon();
	eth->CLC.U = 0;
	lock_wdtcon();
	/* read back for activating module */
	(void)eth->CLC.U;

	/* initialise IO pins */
	eth_init_gpio();

	/* first check for complete (power on) reset */
	/* DMA_MODE can only be written if B.SWR == 0 (!) */
	while (eth->DMA_MODE.B.SWR)
		;
	/* reset complete MAC/MTL/DMA */
	eth->DMA_MODE.B.SWR = 1;
	/* wait until reset done */
	while (eth->DMA_MODE.B.SWR)
		;

#if 0
	/* prepare clock setting for MDIO (150 - 250 MHz) */
	eth->MAC_MDIO_ADDRESS.U = (4 << IFX_GETH_MAC_MDIO_ADDRESS_CR_OFF);
#else
	/* prepare clock setting for MDIO (60 - 100 MHz) */
	eth->MAC_MDIO_ADDRESS.U = (0 << IFX_GETH_MAC_MDIO_ADDRESS_CR_OFF);
#endif

	/* short delay after reset */
	/* copy supplied MAC address into local data structure */
	for (i = 0; i < sizeof(emData.etheraddr); ++i)
		emData.etheraddr[i] = eth_mac[i];

	/* write the MAC address */
	eth->MAC_ADDRESS_LOW0.U	 = (unsigned int)eth_mac[0]
							 | ((unsigned int)eth_mac[1] << 8)
							 | ((unsigned int)eth_mac[2] << 16)
							 | ((unsigned int)eth_mac[3] << 24);
	eth->MAC_ADDRESS_HIGH0.U = (unsigned int)eth_mac[4]
							 | ((unsigned int)eth_mac[5] << 8);


	/* prepare MAC and DMA register */
#if 0
	eth->MAC_EXT_CONFIGURATION.U = ETHERMTU | (IFX_GETH_MAC_EXT_CONFIGURATION_DCRCC_MSK << IFX_GETH_MAC_EXT_CONFIGURATION_DCRCC_OFF);
#else
	eth->MAC_EXT_CONFIGURATION.U = (IFX_GETH_MAC_EXT_CONFIGURATION_DCRCC_MSK << IFX_GETH_MAC_EXT_CONFIGURATION_DCRCC_OFF);
#endif
#if (USE_RMII_ONLY > 0)
	/* prepare for 100 Mbit/s Full Duplex */
	/* there must be at least 4 cycles between writes to same register ==> use only ONE write */
	eth->MAC_CONFIGURATION.U |= (1 << IFX_GETH_MAC_CONFIGURATION_FES_OFF)
							  | (1 << IFX_GETH_MAC_CONFIGURATION_PS_OFF)
							  | (1 << IFX_GETH_MAC_CONFIGURATION_DM_OFF);
#else
#if 0
	/* prepare for 1000 Mbit/s Full Duplex */
	eth->MAC_CONFIGURATION.B.DM  = 1;
#else
	/* set after autonegotiation */
	eth->MAC_CONFIGURATION.U = 0;
#endif
#endif /* USE_RMII_ONLY > 0 */
#if 0
	/* ToDo: documentation unclear (?!) */
	eth->MTL_RXQ0.OPERATION_MODE.B.RQS = 8;
	eth->MTL_TXQ0.OPERATION_MODE.B.TQS = 8;
#else
	/* allocate all available memory to queue 0 (sizes from MAC_HW_FEATURE1) */
	/* [RX|TX]FIFOSIZE: size := 2 ^ (n + 7) */
	/* RQS|TQS: n := (size / 256) - 1 */
	eth->MTL_RXQ0.OPERATION_MODE.B.RQS = ((1 << ((eth->MAC_HW_FEATURE1.B.RXFIFOSIZE + 7) - 8)) - 1);
	eth->MTL_TXQ0.OPERATION_MODE.B.TQS = ((1 << ((eth->MAC_HW_FEATURE1.B.TXFIFOSIZE + 7) - 8)) - 1);
#endif
	eth->MTL_RXQ0.OPERATION_MODE.B.RSF = 1;
	eth->MTL_TXQ0.OPERATION_MODE.B.TSF = 1;

#if 1
	/* add some additional delay - test for DP83867 */
	HAL_DELAY_US(PHY_TEST_TIME);
	HAL_DELAY_US(PHY_TEST_TIME);
	HAL_DELAY_US(PHY_TEST_TIME);
#endif

	/* Configure the PHY via the RMII|MII interface of the MAC */
	emData.phy_ready = (eth_init_phy(&emData) != 0) ? 0 : 1;

	/* Configure RX and TX DMA engines and descriptor arrays */
	eth_init_memory();

#if (USE_RMII_ONLY == 0)
	/* without Receive All (RA) or Promiscuous Mode (PR) no unicast frames come in (??) */
#if 0
	eth->MAC_PACKET_FILTER.U = (IFX_GETH_MAC_PACKET_FILTER_RA_MSK << IFX_GETH_MAC_PACKET_FILTER_RA_OFF);
#else
	eth->MAC_PACKET_FILTER.U = (IFX_GETH_MAC_PACKET_FILTER_PR_MSK << IFX_GETH_MAC_PACKET_FILTER_PR_OFF);
#endif
#endif /* USE_RMII_ONLY == 0 */

	/* Enable the receive and transmit data paths */
	eth->MTL_TXQ0.OPERATION_MODE.B.TXQEN = 2;
	eth->MAC_RXQ_CTRL0.B.RXQ0EN = 2;
	/* there must be at least 4 cycles between writes to same register ==> use only ONE write */
	eth->MAC_CONFIGURATION.U |= (1 << IFX_GETH_MAC_CONFIGURATION_RE_OFF) | (1 << IFX_GETH_MAC_CONFIGURATION_TE_OFF);
	eth->DMA_CH[0].RX_CONTROL.B.SR = 1;
	eth->DMA_CH[0].TX_CONTROL.B.ST = 1;

	return emData.phy_ready ? 0 : -1;
}

/* returns 1 if ethernet needs some action, else 0 */
int eth_tc3x_poll(void)
{
	int ret;
	unsigned int status = eth->DMA_CH[0].STATUS.U;

	if (status & (IFX_GETH_DMA_CH_STATUS_RBU_MSK << IFX_GETH_DMA_CH_STATUS_RBU_OFF))
	{
		/* RX Buffer Unavailable: clear condition and resume */
		eth->DMA_CH[0].STATUS.U = (IFX_GETH_DMA_CH_STATUS_RBU_MSK << IFX_GETH_DMA_CH_STATUS_RBU_OFF);
		/* inform DMA about change */
		eth->DMA_CH[0].RXDESC_TAIL_POINTER.U = (unsigned int)&eth_Memory->rxdesc[ETH_MAX_BUFFER - 1];
	}
	if (status & (IFX_GETH_DMA_CH_STATUS_RPS_MSK << IFX_GETH_DMA_CH_STATUS_RPS_OFF))
	{
		/* RX suspended: clear condition and resume */
		eth->DMA_CH[0].STATUS.U = (IFX_GETH_DMA_CH_STATUS_RPS_MSK << IFX_GETH_DMA_CH_STATUS_RPS_OFF);
		/* inform DMA about change */
		eth->DMA_CH[0].RXDESC_TAIL_POINTER.U = (unsigned int)&eth_Memory->rxdesc[ETH_MAX_BUFFER - 1];
	}

	if (status & (IFX_GETH_DMA_CH_STATUS_TBU_MSK << IFX_GETH_DMA_CH_STATUS_TBU_OFF))
	{
		/* TX Buffer Unavailable: clear condition and resume */
		eth->DMA_CH[0].STATUS.U = (IFX_GETH_DMA_CH_STATUS_TBU_MSK << IFX_GETH_DMA_CH_STATUS_TBU_OFF);
		/* inform DMA about change */
		eth->DMA_CH[0].TXDESC_TAIL_POINTER.U = (unsigned int)&eth_Memory->txdesc[ETH_MAX_BUFFER - 1];
	}
	if (status & (IFX_GETH_DMA_CH_STATUS_TPS_MSK << IFX_GETH_DMA_CH_STATUS_TPS_OFF))
	{
		/* TX suspended: clear condition and resume */
		eth->DMA_CH[0].STATUS.U = (IFX_GETH_DMA_CH_STATUS_TPS_MSK << IFX_GETH_DMA_CH_STATUS_TPS_OFF);
		/* inform DMA about change */
		eth->DMA_CH[0].TXDESC_TAIL_POINTER.U = (unsigned int)&eth_Memory->txdesc[ETH_MAX_BUFFER - 1];
	}
	if (status & (IFX_GETH_DMA_CH_STATUS_TI_MSK << IFX_GETH_DMA_CH_STATUS_TI_OFF))
	{
		/* switch to next descriptor */
		if (eth_Memory->currTx == eth_Memory->lastTx)
		{
			eth_Memory->currTx = &eth_Memory->txdesc[0];
		}
		else
		{
			eth_Memory->currTx = &eth_Memory->currTx[1];
		}
		/* clear status */
		eth->DMA_CH[0].STATUS.U = (IFX_GETH_DMA_CH_STATUS_TI_MSK << IFX_GETH_DMA_CH_STATUS_TI_OFF);
		/* inform DMA about change */
		eth->DMA_CH[0].TXDESC_TAIL_POINTER.U = (unsigned int)&eth_Memory->txdesc[ETH_MAX_BUFFER - 1];
	}

	ret = (status & (IFX_GETH_DMA_CH_STATUS_RI_MSK << IFX_GETH_DMA_CH_STATUS_RI_OFF)) ? 1 : 0;

	if (ret)
	{
		/* check next descriptor */
		eth_rxdesc_t *pRx = eth_Memory->currRx;
		if (pRx == eth_Memory->lastRx)
		{
			pRx = &eth_Memory->rxdesc[0];
		}
		else
		{
			pRx = &pRx[1];
		}
		/* check if there is another frame pending */
		if (1 == pRx->rdes3.rd.OWN)
		{
			/* clear RX status flag if this one is the only one */
			eth->DMA_CH[0].STATUS.U = (IFX_GETH_DMA_CH_STATUS_RI_MSK << IFX_GETH_DMA_CH_STATUS_RI_OFF);
		}
	}

	return ret;
}

/* returns number of bytes send or < 0 */
int eth_tc3x_send(const char *buf, int len)
{
	int res = -1;

	/* check link state of network connection */
	emData.phy_ready = (eth_phy_check_link(emData.phy_addr) != 0) ? 0 : 1;

	if (emData.phy_ready)
	{
		eth_txdesc_t *pTx = eth_Memory->headTx;

		/* check for place in descriptor list */
		if (0 == pTx->tdes3.rd.OWN)
		{
			memcpy(pTx->tdes0.buffer1, buf, (size_t)len);
			pTx->tdes2.rd.TBS1 = (unsigned int)len;
			pTx->tdes3.rd.FL_TPL = (unsigned int)len;
			pTx->tdes3.rd.CIC = 3;
			pTx->tdes3.rd.FD = 1;
			pTx->tdes3.rd.LD = 1;
			pTx->tdes2.rd.IOC = 1;
			pTx->tdes3.rd.OWN = 1;
			res = len;
			/* clear any outstanding TX flags */
			eth->DMA_CH[0].STATUS.U = (IFX_GETH_DMA_CH_STATUS_TBU_MSK << IFX_GETH_DMA_CH_STATUS_TBU_OFF)
									| (IFX_GETH_DMA_CH_STATUS_TPS_MSK << IFX_GETH_DMA_CH_STATUS_TPS_OFF)
									| (IFX_GETH_DMA_CH_STATUS_ETI_MSK << IFX_GETH_DMA_CH_STATUS_ETI_OFF);
			/* switch to next descriptor for application usage */
			if (pTx == eth_Memory->lastTx)
			{
				eth_Memory->headTx = &eth_Memory->txdesc[0];
			}
			else
			{
				eth_Memory->headTx = &eth_Memory->headTx[1];
			}
			/* inform DMA about change */
			eth->DMA_CH[0].TXDESC_TAIL_POINTER.U = (unsigned int)&eth_Memory->txdesc[ETH_MAX_BUFFER - 1];
		}
	}

	return res;
}

/* returns number of bytes read or < 0 */
int eth_tc3x_read(char *buf, int bufsize)
{
	unsigned int byteCount;
	int res = -1;

	/* process next buffered frame ... if there is one */
	if (0 == eth_Memory->currRx->rdes3.rd.OWN)
	{

		byteCount = eth_Memory->currRx->rdes3.wb.PL;
		/* check for buffer size */
		if (byteCount > (unsigned int)bufsize)
		{
			byteCount = (unsigned int)bufsize;
		}

		/* copy received data into supplied buffer */
		memcpy(buf, eth_Memory->currRx->rdes0.buffer1, byteCount);

		/* read successful */
		res = (int)byteCount;

		/* prepare this descriptor for reuse by DMA */
#if 0
		/* clear any error flags */
		eth_Memory->currRx->rdes2.u = 0;
#endif
		eth_Memory->currRx->rdes3.u = 0;
		eth_Memory->currRx->rdes3.rd.BUF1V = 1;
		eth_Memory->currRx->rdes3.rd.IOC = 1;
		eth_Memory->currRx->rdes3.rd.OWN = 1;
		/* switch to next descriptor */
		if (eth_Memory->currRx == eth_Memory->lastRx)
		{
			eth_Memory->currRx = &eth_Memory->rxdesc[0];
		}
		else
		{
			eth_Memory->currRx = &eth_Memory->currRx[1];
		}
		/* inform DMA about change */
		eth->DMA_CH[0].RXDESC_TAIL_POINTER.U = (unsigned int)&eth_Memory->rxdesc[ETH_MAX_BUFFER - 1];
	}

	return res;
}

/* shutdown ethernet device */
void eth_tc3x_stop(int wait)
{
	if (wait)
	{
		/* wait until last TX is finished (ACK for last HtcBootCtrl command) */
		while (0 == (eth->DMA_CH[0].STATUS.U & (IFX_GETH_DMA_CH_STATUS_TI_MSK << IFX_GETH_DMA_CH_STATUS_TI_OFF)))
		{
			;
		}
	}

	/* Stop DMA transfers */
	eth->DMA_CH[0].RX_CONTROL.B.SR = 0;
	eth->DMA_CH[0].TX_CONTROL.B.ST = 0;
	/* Disable the receive and transmit data paths */
	/* there must be at least 4 cycles between writes to same register ==> use only ONE write */
	eth->MAC_CONFIGURATION.U &= ~((1 << IFX_GETH_MAC_CONFIGURATION_RE_OFF) | (1 << IFX_GETH_MAC_CONFIGURATION_TE_OFF));
}
