/*====================================================================
* Project:  Board Support Package (BSP)
* Function: Transmit and receive characters via TriCore's serial line.
*           (Infineon controller TC39xx)
*           (polling variant)
*
* Copyright HighTec EDV-Systeme GmbH 1982-2020
*====================================================================*/

#include <machine/wdtcon.h>
#include "rs232.h"

#include "tc_inc_path.h"

#include TC_INCLUDE(TCPATH/IfxPort_reg.h)
#include TC_INCLUDE(TCPATH/IfxAsclin_reg.h)

#ifndef BAUDRATE
#define BAUDRATE		38400
#endif /* BAUDRATE */


/* Port Modes */
#define IN_NOPULL0		0x00	/* Port Input No Pull Device */
#define IN_PULLDOWN		0x01	/* Port Input Pull Down Device */
#define IN_PULLUP		0x02	/* Port Input Pull Up Device */
#define IN_NOPULL3		0x03	/* Port Input No Pull Device */
#define OUT_PPGPIO		0x10	/* Port Output General Purpose Push/Pull */
#define OUT_PPALT1		0x11	/* Port Output Alternate 1 Function Push/Pull */
#define OUT_PPALT2		0x12	/* Port Output Alternate 2 Function Push/Pull */
#define OUT_PPALT3		0x13	/* Port Output Alternate 3 Function Push/Pull */
#define OUT_PPALT4		0x14	/* Port Output Alternate 4 Function Push/Pull */
#define OUT_PPALT5		0x15	/* Port Output Alternate 5 Function Push/Pull */
#define OUT_PPALT6		0x16	/* Port Output Alternate 6 Function Push/Pull */
#define OUT_PPALT7		0x17	/* Port Output Alternate 7 Function Push/Pull */
#define OUT_ODGPIO		0x18	/* Port Output General Purpose Open Drain */
#define OUT_ODALT1		0x19	/* Port Output Alternate 1 Function Open Drain */
#define OUT_ODALT2		0x1A	/* Port Output Alternate 2 Function Open Drain */
#define OUT_ODALT3		0x1B	/* Port Output Alternate 3 Function Open Drain */
#define OUT_ODALT4		0x1C	/* Port Output Alternate 4 Function Open Drain */
#define OUT_ODALT5		0x1D	/* Port Output Alternate 5 Function Open Drain */
#define OUT_ODALT6		0x1E	/* Port Output Alternate 6 Function Open Drain */
#define OUT_ODALT7		0x1F	/* Port Output Alternate 7 Function Open Drain */


static Ifx_ASCLIN * const asc  = (Ifx_ASCLIN *)&MODULE_ASCLIN0;
#if (RUN_ON_PHYCORE > 0) && defined(USE_ONBOARD_WIGGLER)
static Ifx_P      * const port = (Ifx_P *)&MODULE_P15;
#else
static Ifx_P      * const port = (Ifx_P *)&MODULE_P14;
#endif /* (RUN_ON_PHYCORE > 0) && USE_ONBOARD_WIGGLER */

/* definitions for RX error conditions */
#define ASC_ERROR_MASK		((1u << 16) | (1u << 18) | (1u << 26))	/* PE | FE | RFO */
#define ASC_CLRERR_MASK		((1u << 16) | (1u << 18) | (1u << 26))	/* PEC | FEC | RFOC */

/* UART primitives */
#define PUT_CHAR(u, c)		((u)->TXDATA.U = (c))
#define GET_CHAR(u)			((u)->RXDATA.U)
#define RX_READY(u)			((u)->FLAGS.B.RFL != 0)				/* Receive FIFO Level */
#define RX_CLEAR(u)			((u)->FLAGSCLEAR.U = (1u << 28))	/* FLAGSCLEAR.RFLC */
#define TX_READY(u)			((u)->FLAGS.B.TFL != 0)				/* Transmit FIFO Level */
#define TX_CLEAR(u)			((u)->FLAGSCLEAR.U = (1u << 31))	/* FLAGSCLEAR.TFLC */
#define START_TX(u)			((u)->FLAGSSET.U   = (1u << 31))	/* FLAGSSET.TFLS */
#define GET_ERROR_STATUS(u)	(((u)->FLAGS.U) & ASC_ERROR_MASK)
#define RESET_ERROR(u)		((u)->FLAGSCLEAR.U = ASC_CLRERR_MASK)


/* Initialise asynchronous interface to operate at BAUDRATE,8,n,1 */
void _init_uart(int baudrate)
{
	unsigned int numerator;
	unsigned int denominator;

#if (RUN_ON_PHYCORE > 0) && defined(USE_ONBOARD_WIGGLER)
	/* on board wiggler is connected to ASCLIN0 */
	/* ARX0B/P15.3 (RXD), ATX0/P15.2 (TXD) */
	/* Set TXD/P15.2 to "output" and "high" */
	port->IOCR0.B.PC2 = OUT_PPALT2;
	port->OMR.B.PS2 = 1;
#else
	/* on board wiggler is connected to ASCLIN0 */
	/* ARX0A/P14.1 (RXD), ATX0/P14.0 (TXD) */
	/* Set TXD/P14.0 to "output" and "high" */
	port->IOCR0.B.PC0 = OUT_PPALT2;
	port->OMR.B.PS0 = 1;
#endif /* (RUN_ON_PHYCORE > 0) && USE_ONBOARD_WIGGLER */

	if (baudrate <= 0)
		baudrate = BAUDRATE;

	/* baudrate values at 100 MHz */
	denominator = 3125;
	switch (baudrate)
	{
		case 9600 :
			numerator = 48;
			break;
		case 19200 :
			numerator = 48 * 2;
			break;
		case 38400 :
		default :
			numerator = 48 * 4;
			break;
		case 57600 :
			numerator = 48 * 6;
			break;
		case 115200 :
			numerator = 48 * 12;
			break;
	}

	/* Enable ASC0 */
	unlock_wdtcon();
	asc->CLC.U = 0;
	/* TXD and RXD pins: automotive level, fast */
#if (RUN_ON_PHYCORE > 0) && defined(USE_ONBOARD_WIGGLER)
	port->PDR0.B.PD2 = 0;
	port->PDR0.B.PD3 = 0;
#else
	port->PDR0.B.PD0 = 0;
	port->PDR0.B.PD1 = 0;
#endif /* (RUN_ON_PHYCORE > 0) && USE_ONBOARD_WIGGLER */
	lock_wdtcon();
	/* read back for activating module */
	(void)asc->CLC.U;

	/* Program ASC0 */
	asc->CSR.U = 0;

#if (RUN_ON_PHYCORE > 0) && defined(USE_ONBOARD_WIGGLER)
	/* select ARX0B/P15.3 as input pin (needs CSR.CLKSEL == 0) */
	asc->IOCR.B.ALTI = 1;
#else
	/* select ARX0A/P14.1 as input pin (needs CSR.CLKSEL == 0) */
	asc->IOCR.B.ALTI = 0;
#endif /* (RUN_ON_PHYCORE > 0) && USE_ONBOARD_WIGGLER */

	/* configure TX and RX FIFOs */
	asc->TXFIFOCON.U = (1 << 6)		/* INW: (1 == 1 byte) */
					 | (1 << 1)		/* ENO */
					 | (1 << 0);	/* FLUSH */
	asc->RXFIFOCON.U = (1 << 6)		/* OUTW: (1 == 1 byte) */
					 | (1 << 1)		/* ENI */
					 | (1 << 0);	/* FLUSH */

	asc->BITCON.U = ( 9 << 0)		/* PRESCALER: 10 */
				  | (15 << 16)		/* OVERSAMPLING: 16 */
				  | ( 9 << 24)		/* SAMPLEPOINT: position 7,8,9 */
				  | (1u << 31);		/* SM: 3 samples per bit */

	/* data format: 8N1 */
	asc->FRAMECON.U = (1 << 9)		/* STOP: 1 bit */
					| (0 << 16)		/* MODE: Init */
					| (0 << 30);	/* PEN: no parity */
	asc->DATCON.U = (7 << 0);		/* DATLEN: 8 bit */

	/* set baudrate value */
	asc->BRG.U = (denominator << 0)	/* DENOMINATOR */
			   | (numerator << 16);	/* NUMERATOR */

	asc->FRAMECON.B.MODE = 1;		/* ASC mode */
	asc->CSR.U = 1;					/* select CLC as clock source */

	START_TX(asc);
}

/* Send character CHR via the serial line */
void _out_uart(const unsigned char chr)
{
	/* wait until space is available in the FIFO */
	while (!TX_READY(asc))
		;

	TX_CLEAR(asc);

	/* send the character */
	PUT_CHAR(asc, (unsigned int)chr);
}

/* Receive (and wait for) a character from the serial line */
unsigned char _in_uart(void)
{
	unsigned char ch;

	/* poll until a character is available */
	while (0 == _poll_uart(&ch))
	{
		;
	}

	return ch;
}

/* Check the serial line if a character has been received.
   returns 1 and the character in *chr if there is one
   else 0
 */
int _poll_uart(unsigned char *chr)
{
	int res = 0;

	if (RX_READY(asc))
	{
		*chr = (unsigned char)GET_CHAR(asc);
		RX_CLEAR(asc);
		/* check for error condition */
		if (GET_ERROR_STATUS(asc))
		{
			/* reset error flags */
			RESET_ERROR(asc);
		}
		else
		{
			res = 1;
		}
	}

	return res;
}
