/*====================================================================
* Project:  Board Support Package (BSP)
* Function: Serial I/O for stdin, stdout and stderr
*           POSIX read/write functions
*
* Copyright HighTec EDV-Systeme GmbH 1982-2020
*====================================================================*/

#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <unistd.h>

#ifndef NON_BLOCKING_SERIALIO
#define NON_BLOCKING_SERIALIO	0
#endif /* NON_BLOCKING_SERIALIO */

#include "rs232.h"
#include "serialio.h"

extern int fileno(FILE *);

/* Initialise serial communication to operate at BAUDRATE,8,n,1 */
void serialio_init(int baudrate)
{
#if (NON_BLOCKING_SERIALIO > 0)
	setvbuf(stdout, NULL, _IOLBF, 0);
	setvbuf(stderr, NULL, _IOLBF, 0);
#endif /* NON_BLOCKING_SERIALIO */
	_init_uart(baudrate);
}

/* Finish serial communication */
void serialio_exit(void)
{
#if (NON_BLOCKING_SERIALIO > 0)
	/* wait until sending has finished */
	while (_uart_sending())
		;
#endif /* NON_BLOCKING_SERIALIO */
}

/* Newlib has this definition for the return type. */
#ifndef _READ_WRITE_RETURN_TYPE
#define _READ_WRITE_RETURN_TYPE size_t
#endif

/* POSIX read function */
/* read characters from file descriptor fd into given buffer, at most count bytes */
/* returns number of characters in buffer */
_READ_WRITE_RETURN_TYPE read(int fd, void *buffer, size_t count)
{
	size_t index = 0;

	if (fileno(stdin) == fd)
	{
#if (NON_BLOCKING_SERIALIO > 0)
		char *ptr = (char *)buffer;
		do
		{
			if (1 == _uart_getchar(ptr))
			{
				++ptr;
				++index;
			}
			else
			{
				/* wait at least for 1 character */
				if (index >= 1)
				{
					break;
				}
			}
		} while (index < count);
#else
		unsigned char *ptr = (unsigned char *)buffer;
		do
		{
			if (1 == _poll_uart(ptr))
			{
				++ptr;
				++index;
			}
			else
			{
				/* wait at least for 1 character */
				if (index >= 1)
				{
					break;
				}
			}
		} while (index < count);
#endif /* NON_BLOCKING_SERIALIO */
	}

	return index;
}

/* POSIX write function */
/* write content of buffer to file descriptor fd */
/* returns number of characters that have been written */
_READ_WRITE_RETURN_TYPE write(int fd, const void *buffer, size_t count)
{
	size_t index = 0;

	if ((fileno(stdout) == fd) || (fileno(stderr) == fd))
	{
#if (NON_BLOCKING_SERIALIO > 0)
		int ret = _uart_send((const char *)buffer, (int)count);
		if (ret)
		{
			index = count;
		}
#else
		const unsigned char *ptr = (const unsigned char *)buffer;
		while (index < count)
		{
			_out_uart(*ptr++);
			++index;
		}
#endif /* NON_BLOCKING_SERIALIO */
	}

	return index;
}
