/*====================================================================
* Project:  Board Support Package
* Function: test for using W5300 ethernet
*           [test version for phyCORE-TC39xB & phyCORE-TC29xB]
*
* Copyright HighTec EDV-Systeme GmbH 1982-2019
*====================================================================*/

#include <stdio.h>
#include <string.h>

#include "system.h"
#include "lanw5300.h"

#define IOBASE			0xA5000000
#define ETHMAC			"\x0\x50\xC2\xA5\xB4\x81"
#define MY_IP			"\xC0\xA8\x0\xC7"			/* 192.168.0.199 */

/* definition of constants: no need of ntohs() for a little endian target */
#define FRAME_IP		0x0008
#define FRAME_ARP		0x0608
#define OP_ARP_REQ		0x0100
#define OP_ARP_ANS		0x0200
#define HARDW_ETH10		0x0100	/* hardware-type 10Mbps Ethernet */
#define IP_HLEN_PLEN	0x0406	/* MAC = 6 byte long, IP = 4 byte long */
#define IP_VER_IHL		0x0045	/* IPv4, Header Length = 5x32 bit */
#define PROT_ICMP		0x0100	/* internet Control Message Protocol */
#define ICMP_ECHO		0x0008	/* message is an echo request */
#define ICMP_ECHO_REPLY	0x0000	/* message is an echo reply */

#define MAXBUFSIZE	2048

static unsigned char rxbuf[MAXBUFSIZE];
static unsigned char txbuf[MAXBUFSIZE];

static unsigned int rxcnt = 0;
static unsigned int txcnt = 0;
static unsigned int rxerr = 0;
static unsigned int txerr = 0;

static __inline__ unsigned short ntohs(unsigned short Data)
{
	return (Data >> 8) | (Data << 8);
}

static unsigned short CalcChecksum(void *Start, unsigned int Count)
{
	unsigned long Sum = 0;
	unsigned short *piStart = (unsigned short *)Start;

	while (Count > 1)							/* sum words */
	{
		Sum += *piStart++;
		Count -= 2;
	}

	if (Count)									/* add left-over byte, if any */
		Sum += *(unsigned char *)piStart;

	while (Sum >> 16)							/* fold 32-bit sum to 16 bits */
		Sum = (Sum & 0xFFFF) + (Sum >> 16);

	return (unsigned short)~Sum;
}

static int processFrame(void)
{
	size_t tosend = 0;

	if ((*(unsigned short *)(rxbuf+12) == FRAME_ARP)
	    && (*(unsigned short *)(rxbuf+20) == OP_ARP_REQ)
	    && !memcmp(MY_IP, rxbuf+38, 4))
	{
		/* reply to an ARP request */
		/* ethernet header */
		memcpy(txbuf, rxbuf+6, 6);
		memcpy(txbuf+6, ETHMAC, 6);
		*(unsigned short *)(txbuf+12) = FRAME_ARP;

		/* ARP entry */
		*(unsigned short *)(txbuf+14) = HARDW_ETH10;
		*(unsigned short *)(txbuf+16) = FRAME_IP;
		*(unsigned short *)(txbuf+18) = IP_HLEN_PLEN;
		*(unsigned short *)(txbuf+20) = OP_ARP_ANS;
		memcpy(txbuf+22, ETHMAC, 6);
		memcpy(txbuf+28, MY_IP, 4);
		memcpy(txbuf+32, rxbuf+6, 6);
		memcpy(txbuf+38, rxbuf+28, 4);

		/* send ARP answer */
		tosend = 42;
	}
	else
	if ((*(unsigned short *)(rxbuf+12) == FRAME_IP)
	    && ((*(unsigned short *)(rxbuf+14) & 0x00FF) == IP_VER_IHL)
	    && ((*(unsigned short *)(rxbuf+22) & 0xFF00) == PROT_ICMP)
	    && !memcmp(MY_IP, rxbuf+30, 4)
	    && ((*(unsigned short *)(rxbuf+34) & 0x00FF) == ICMP_ECHO))
	{
		unsigned short dataCnt = ntohs(*(unsigned short *)(rxbuf+16));

		/* reply to an ICMP_ECHO request */
		/* ethernet header */
		memcpy(txbuf, rxbuf+6, 6);
		memcpy(txbuf+6, ETHMAC, 6);
		*(unsigned short *)(txbuf+12) = FRAME_IP;

		/* IP header */
		*(unsigned short *)(txbuf+14) = IP_VER_IHL;
		*(unsigned short *)(txbuf+16) = ntohs(dataCnt);
		*(unsigned short *)(txbuf+18) = 0;
		*(unsigned short *)(txbuf+20) = 0;
		*(unsigned short *)(txbuf+22) = 0x0140;	/* protocol + TTL (time to live) */
		*(unsigned short *)(txbuf+24) = 0;		/* head checksum */
		memcpy(txbuf+26, MY_IP, 4);
		memcpy(txbuf+30, rxbuf+26, 4);
		*(unsigned short *)(txbuf+24) = CalcChecksum(txbuf+14, 20);	/* head checksum */

		/* ICMP field */
		*(unsigned short *)(txbuf+34) = ICMP_ECHO_REPLY;
		*(unsigned short *)(txbuf+36) = 0;		/* ICMP message checksum */
		memcpy(txbuf+38, rxbuf+38, dataCnt-24);
		*(unsigned short *)(txbuf+36) = CalcChecksum(txbuf+34, dataCnt-20);	/* ICMP message checksum */

		/* send ICMP echo reply */
		tosend = dataCnt + 14;
	}
	if (tosend)
	{
		if (lanw5300_can_send())
		{
			if (lanw5300_send((const char *)txbuf, tosend) != (int)tosend)
			{
				++txerr;
			}
			++txcnt;
		}
	}

	return 0;
}


#define TEST_EXTINT		1

#ifdef TEST_EXTINT
#include "tc_inc_path.h"

#include TC_INCLUDE(TCPATH/IfxPort_reg.h)

/* external interrupt of LAN chip W5300 on P33.7 */
static Ifx_P      * const port = (Ifx_P *)&MODULE_P33;

#define MAXTESTSIZE	0x1000

static unsigned int valArray[MAXTESTSIZE];

static unsigned int val = 1, valOld = 0, valIndex = 0;

static void check_port(void)
{
	val = (unsigned int)port->IN.U;
	if (val != valOld)
	{
		valArray[valIndex] = val;
		valOld = val;
		if (valIndex < MAXTESTSIZE)
		{
			++valIndex;
		}
	}
}
#endif /* TEST_EXTINT */

int main(void)
{
	int err;

	SYSTEM_Init();

	err = lanw5300_init(IOBASE, (unsigned char *)ETHMAC);

	if (!err)
	{
		while (!err)
		{
#ifdef TEST_EXTINT
			check_port();
#endif /* TEST_EXTINT */
			if (lanw5300_poll())
			{
#ifdef TEST_EXTINT
				check_port();
#endif /* TEST_EXTINT */
				err = (lanw5300_read((char *)rxbuf, sizeof(rxbuf)) <= 0);
				if (!err)
				{
					++rxcnt;
					err = processFrame();
				}
				else
				{
					++rxerr;
				}
			}
		}
	}

	return 0;
}
