/*====================================================================
* Project:  Board Support Package
* Function: Standalone driver for using WIZnet W5300 Ethernet chip
*           [test version for phyCORE-TC39xB & phyCORE-TC29xB]
*
* Copyright HighTec EDV-Systeme GmbH 1982-2019
*====================================================================*/

#include "lanw5300.h"


/* no address translation on phyCORE-TC39xB needed */
#define REAL_ADDRESS(x)	((x))

/* Macros and inline-functions for accessing Ethernet chip */

#define HAL_READ_UINT16(_register_, _value_)	\
		((_value_) = *((volatile unsigned short *)(_register_)))

#define HAL_WRITE_UINT16(_register_, _value_)	\
		(*((volatile unsigned short *)(_register_)) = (_value_))

/* Read a register */
static __inline__ unsigned short get_reg(unsigned char *base, int regno)
{
	unsigned short val;

	HAL_READ_UINT16(base + REAL_ADDRESS(regno), val);

	return val;
}

/* Write a register */
static __inline__ void put_reg(unsigned char *base, int regno, unsigned short val)
{
	HAL_WRITE_UINT16(base + REAL_ADDRESS(regno), val);
}

static __inline__ void put_reg32(unsigned char *base, int regno, unsigned int val)
{
	HAL_WRITE_UINT16(base + REAL_ADDRESS(regno), (unsigned short)(val >> 16));
	HAL_WRITE_UINT16(base + REAL_ADDRESS(regno + 2), (unsigned short)val);
}

static __inline__ unsigned int get_reg32(unsigned char *base, int regno)
{
	unsigned int val;
	unsigned short temp;

	HAL_READ_UINT16(base + REAL_ADDRESS(regno), temp);
	val = (unsigned int)temp << 16;
	HAL_READ_UINT16(base + REAL_ADDRESS(regno + 2), temp);
	val |= temp;

	return val;
}

static __inline__ unsigned short ntohs(unsigned short Data)
{
	return (Data >> 8) | (Data << 8);
}


/* Base address for memory mapped access to the LAN controller */
static unsigned char	*ownBase;
static int resetCnt = 0;
static int openCnt = 0;

/* memory configuration for W5300: we use only SOCKET0 in MACRAW mode */
static const unsigned char w5300_rxbuf_conf[W5300_MAX_SOCK_NUM] = {64, 0, 0, 0, 0, 0, 0, 0};
static const unsigned char w5300_txbuf_conf[W5300_MAX_SOCK_NUM] = {64, 0, 0, 0, 0, 0, 0, 0};

#define CHIP_ID_W5300	0x5300

int lanw5300_init(unsigned int eth_base, unsigned char *eth_mac)
{
	unsigned short val;
	int i;
	int txbuf_total = 0;

	ownBase = (unsigned char *)eth_base;

	/* probe chip by reading the signature in Chip ID register */
	val = get_reg(ownBase, W5300_IDR);

	if (val != CHIP_ID_W5300) return -1;

	/* Reset chip */
	put_reg(ownBase, W5300_MR, W5300_MR_RST);

	/* we need a delay loop here !! */
	for (i = 1000000; i > 0; --i)
	{
		;
	}

	/* wait until RESET finished */
	while (get_reg(ownBase, W5300_MR) & W5300_MR_RST)
	{
		++resetCnt;
	}

	val = get_reg(ownBase, W5300_MR);
#if 1
	/* disable automatic ping, FIFO swap */
	val |= W5300_MR_PB | W5300_MR_FS;
#else
	/* disable automatic ping */
	val |= W5300_MR_PB;
#endif
	put_reg(ownBase, W5300_MR, val);

	/* write the MAC address */
	for (i = 0; i < 6; i += 2)
	{
		put_reg(ownBase, W5300_SHAR + i, ((unsigned short)eth_mac[i] << 8) | (unsigned short)eth_mac[i+1]);
	}

	/* configure memory */
	for (i = 0; i < W5300_MAX_SOCK_NUM; i += 2)
	{
		put_reg(ownBase, W5300_TMSR_01 + i, ((unsigned short)w5300_txbuf_conf[i] << 8) | (unsigned short)w5300_txbuf_conf[i+1]);
		put_reg(ownBase, W5300_RMSR_01 + i, ((unsigned short)w5300_rxbuf_conf[i] << 8) | (unsigned short)w5300_rxbuf_conf[i+1]);
	}
	/* count total TX memory */
	for (i = 0; i < W5300_MAX_SOCK_NUM; ++i)
	{
		txbuf_total += w5300_txbuf_conf[i];
	}
	/* set FIFO Memory Type (TX + RX) in 8 kB blocks */
	for (i = 0, val = 0; i < txbuf_total / 8; ++i)
	{
		val <<= 1;
		val  |= 1;
	}
	put_reg(ownBase, W5300_MTYPER, val);

	/* mask all interrupts */
	put_reg(ownBase, W5300_IMR, 0);

	/* enable interrupt for SOCKET0 */
	put_reg(ownBase, W5300_IMR, W5300_IR_SINT(0));

	/* open SOCKET0 in MACRAW mode */
	put_reg(ownBase, W5300_Sn_MR(0), W5300_SnMR_MACRAW | W5300_SnMR_MF);
	/* ? must we set a port before the OPEN command ? */
	put_reg(ownBase, W5300_Sn_CR(0), W5300_SnCR_OPEN);

	/* wait until SOCKET is open */
	do
	{
		val = get_reg(ownBase, W5300_Sn_SSR(0)) & 0xFF;
		++openCnt;
	} while (val != W5300_SOCK_MACRAW);

	/* enable RX and TX interrupts on SOCKET0 (?) */
	put_reg(ownBase, W5300_Sn_IMR(0), W5300_SnIR_RECV | W5300_SnIR_SENDOK);

	return 0;
}

static int sending = 0;

/*
	check if W5300 is ready to accept the
	frame we want to send
 */

int lanw5300_can_send(void)
{
	return (sending == 0);
}

int lanw5300_poll(void)
{
	int ret = 0;
	unsigned short event = 0, ints;

	/* get the (unmasked) requests */
	ints = get_reg(ownBase, W5300_IR);

	/* check SOCKET0 events */
	if (ints & W5300_IR_SINT(0))
	{
		/* get the (unmasked) requests */
		event = get_reg(ownBase, W5300_Sn_IR(0));
		if (event & W5300_SnIR_SENDOK)
		{
			/* mark as ready for next send order */
			sending = 0;
			/* acknowledge this event */
			put_reg(ownBase, W5300_Sn_IR(0), W5300_SnIR_SENDOK);
		}

		if (event & W5300_SnIR_RECV)
		{
			/* read size of data in RX buffer */
			/* necessary for 10BaseT connection */
			int rxsize = (int)get_reg32(ownBase, W5300_Sn_RX_RSR(0));
			if (rxsize > 0)
			{
				ret = 1;
			}
		}
	}

	return ret;
}

int lanw5300_send(const char *buf, int len)
{
	int i = 0;
	int freesize;
	unsigned short status;
	unsigned short *sdata = (unsigned short *)buf;

	if (sending)
	{
		return -1;
	}

	status = get_reg(ownBase, W5300_Sn_SSR(0)) & 0xFF;
	if (status != W5300_SOCK_MACRAW)
	{
		return -1;
	}

	freesize = (int)get_reg32(ownBase, W5300_Sn_TX_FSR(0));
	if (len > freesize)
	{
		return -1;
	}

	/* Put data into SOCKET0 FIFO */
	for (i = 0; i < len; i += sizeof(*sdata))
	{
		put_reg(ownBase, W5300_Sn_TX_FIFOR(0), *sdata++);
	}

	/* inform W5300 about packet size */
	put_reg32(ownBase, W5300_Sn_TX_WRSR(0), (unsigned int)len);
	/* ... and send it */
//	put_reg(ownBase, W5300_Sn_CR(0), W5300_SnCR_SEND);
	put_reg(ownBase, W5300_Sn_CR(0), W5300_SnCR_SEND_MAC);
	sending = 1;

	return len;
}

int lanw5300_read(char *buf, int bufsize)
{
	int i;
	int rxsize;
	int packetsize = 0;
	unsigned short status;
	unsigned short *sdata = (unsigned short *) buf;

	/* read size of data in RX buffer */
	rxsize = (int)get_reg32(ownBase, W5300_Sn_RX_RSR(0));
	if (rxsize > 0)
	{
		/* read size of packet */
		status = get_reg(ownBase, W5300_Sn_RX_FIFOR(0));
		packetsize = (int)ntohs(status);

		/* check for given buffer size */
		if (packetsize > bufsize)
		{
			packetsize = bufsize;
		}

		/* read in the packet */
		for (i = 0; i < packetsize; i += sizeof(*sdata))
		{
			*sdata++ = get_reg(ownBase, W5300_Sn_RX_FIFOR(0));
		}

		/* skip 4 bytes CRC */
		status = get_reg(ownBase, W5300_Sn_RX_FIFOR(0));
		status = get_reg(ownBase, W5300_Sn_RX_FIFOR(0));

		/* mark end of receive */
		put_reg(ownBase, W5300_Sn_CR(0), W5300_SnCR_RECV);

		/* read RX size again */
		rxsize = (int)get_reg32(ownBase, W5300_Sn_RX_RSR(0));
		/* to cope with 10 Mbit slowness: check for less than header size */
		if (rxsize < 14)
		{
			/* mark end of read */
			put_reg(ownBase, W5300_Sn_IR(0), W5300_SnIR_RECV);
		}
	}

	return packetsize;
}
